from typing import List, Optional, Union
import dataclasses
import json
import math

import numpy as np

from realhf.api.core.dfg import MFCDef
from realhf.api.quickstart.model import ParallelismConfig
from realhf.base.cluster import spec as cluster_spec
from realhf.base.slurm_utils import (
    are_ones_contiguous,
    nodelist_from_nodes,
    parse_nodelist,
)


@dataclasses.dataclass
class DeviceMesh:
    # number of total nodes, n_gpus_per_node=8
    n_nodes: int
    n_gpus_per_node: int
    # a 2D binary array of current device mesh name
    # shape: (n_nodes, n_gpus_per_node)
    mapping: np.ndarray
    # For slurm cluster: nodelist string of all
    # allocated nodes in the cluster
    global_mesh_name: str = None
    # For slurm cluster: nodelist string this device mesh
    name: str = None
    # cluster info, GPU memory cap in bytes
    gpu_memory_capacity: int = 80 * (1024**3)

    def to_dict(self):
        return dict(
            n_nodes=self.n_nodes,
            n_gpus_per_node=self.n_gpus_per_node,
            mapping=self.mapping.tolist(),
            global_mesh_name=self.global_mesh_name,
            name=self.name,
            gpu_memory_capacity=self.gpu_memory_capacity,
        )

    @staticmethod
    def from_dict(d):
        device_mesh = DeviceMesh(**d)
        device_mesh.mapping = np.array(d["mapping"])
        return device_mesh

    def __post_init__(self):
        if self.global_mesh_name is None:
            self.global_mesh_name = (
                f"{cluster_spec.node_name_prefix}[01-{self.n_nodes:02d}]"
                if self.n_nodes > 1
                else f"{cluster_spec.node_name_prefix}01"
            )

        if self.global_mesh_name is not None and self.name is None:
            self.name = device_mesh_name_from_mapping(
                self.global_mesh_name, self.mapping
            )

    def __eq__(self, other: "DeviceMesh"):
        assert (
            self.global_mesh_name is None
            or self.global_mesh_name == other.global_mesh_name
        ), "Only device meshes that on the same cluster mesh is comparable"
        return np.all(self.mapping == other.mapping)

    def __repr__(self):
        return f"DeviceMesh({self.name} in {self.global_mesh_name})"

    def __op_assertion(self, other: "DeviceMesh"):
        assert (
            self.global_mesh_name is None
            or self.global_mesh_name == other.global_mesh_name
        ), "operation only support device meshes on the same cluster nodes"
        assert self.n_nodes == other.n_nodes
        assert self.n_gpus_per_node == other.n_gpus_per_node

    def overlap(self, other: "DeviceMesh") -> bool:
        self.__op_assertion(other)
        return np.any(self.mapping & other.mapping)

    def contain(self, other: "DeviceMesh") -> bool:
        self.__op_assertion(other)
        return np.all(self.mapping & other.mapping == self.mapping)

    def contained_by(self, other: "DeviceMesh") -> bool:
        self.__op_assertion(other)
        return np.all(self.mapping & other.mapping == other.mapping)

    def sub_device_meshes(self, min_n_gpus: int = 4) -> List["DeviceMesh"]:
        """Find sub device meshes of this device mesh with at least min_n_gpus gpus.
        Sub device meshes have following constraints:
            1. Sub device meshes have the same cluster mesh.
            2. Sub device meshes of multiple nodes must contain consecutive nodes
               in the cluster mesh.
            3. Sub device meshes can only be of shape 1x1, 1x2, 1x4, 1x8 or Nx8
            4. If sub device meshes are of shape 1x2 or 1x4, the start GPU id
               must be 0, 2, 4, 6 for 1x2 and 0, 4 for 1x4.
        """
        sub_mappings = []
        rows, cols = np.where(self.mapping == 1)

        unique_rows = np.unique(rows)
        for row in unique_rows:
            this_cols = cols[rows == row]
            # print(row, this_cols)
            assert (
                self.n_gpus_per_node % min_n_gpus == 0
                or min_n_gpus % self.n_gpus_per_node == 0
            )
            n_gpus = min_n_gpus
            while n_gpus < min(self.n_gpus_per_node, np.sum(self.mapping)):
                for start in range(
                    np.min(this_cols), self.n_gpus_per_node, n_gpus
                ):
                    # print(row, start, start+n_gpus)
                    sub_mapping = np.zeros(
                        (self.n_nodes, self.n_gpus_per_node), dtype=np.int32
                    )
                    sub_mapping[row, start : start + n_gpus] = 1
                    sub_mappings.append(sub_mapping)
                n_gpus *= 2

        for n_rows in range(1, len(unique_rows) + 1):
            for start in range(0, len(unique_rows) - n_rows + 1):
                sub_mapping = np.zeros(
                    (self.n_nodes, self.n_gpus_per_node), dtype=np.int32
                )
                sub_mapping[start : start + n_rows, :] = 1
                sub_mappings.append(sub_mapping)

        return [
            DeviceMesh(
                n_nodes=self.n_nodes,
                n_gpus_per_node=self.n_gpus_per_node,
                mapping=sub_mapping,
                global_mesh_name=self.global_mesh_name,
                name=device_mesh_name_from_mapping(
                    self.global_mesh_name, sub_mapping
                ),
            )
            for sub_mapping in sub_mappings
        ]

    def _is_valid_mapping(self) -> bool:
        if self.mapping.shape != (self.n_nodes, self.n_gpus_per_node):
            raise RuntimeError(
                f"Invalid mapping shape {self.mapping.shape} " f"{self.name}"
            )
        if not np.all(np.logical_or(self.mapping == 0, self.mapping == 1)):
            raise RuntimeError(f"Invalid mapping value {self.mapping}")

        assert math.log(self.n_gpus_per_node, 2).is_integer()

        one_node_valid_gpus = [
            2**i for i in range(int(math.log(self.n_gpus_per_node, 2)))
        ]
        if self.mapping.sum() < self.n_gpus_per_node:
            if not any(self.mapping.sum() == g for g in one_node_valid_gpus):
                raise RuntimeError(f"Invalid mapping sum {self.mapping}")
        else:
            if not (
                self.mapping.sum() % self.n_gpus_per_node == 0
                and np.all(
                    np.logical_or(
                        self.mapping.sum(1) == self.n_gpus_per_node,
                        self.mapping.sum(1) == 0,
                    )
                )
            ):
                raise RuntimeError(f"Invalid mapping sum {self.mapping}")
        if not are_ones_contiguous(self.mapping.flatten()):
            raise RuntimeError(
                f"mapping devices are not contiguous {self.mapping}"
            )
        return True


def make_device_mesh_from_name(global_mesh_name: str, name: str):
    """
    DeviceMesh name format: <prefix><node_indices>[:<gpu_ids>]
        slurm_nodelist is the name of slurm nodes the mesh is on, should follow slurm convention,
        for example "NODE[40-43]" or "NODE[01,11,13-14]" with prefix NODE,
        if n_nodes=1, gpu_ids are the gpu id list delimited by comma if n_gpus < 8,
        for example "0,1,2,3" or "0,1". An example of full device mesh name
        in this situation is "NODE40:0,1,2,3"

    Note: cluster device mesh name must occupy entire nodes.
    """
    prefix = cluster_spec.node_name_prefix
    node_list = parse_nodelist(global_mesh_name, prefix)
    n_nodes = len(node_list)
    n_gpus_per_node = 8

    gpu_ids = None
    if ":" in name:
        node_names, gpu_ids = name.split(":")
        gpu_ids = list(map(int, gpu_ids.split(",")))
    else:
        node_names = name
    node_names = parse_nodelist(node_names, prefix)
    mapping = np.zeros((n_nodes, n_gpus_per_node), dtype=np.int32)
    if gpu_ids is None:
        node_indices = [node_list.index(node_name) for node_name in node_names]
        mapping[node_indices, :] = 1
    else:
        assert len(node_names) == 1
        node_index = node_list.index(node_names[0])
        mapping[node_index, gpu_ids] = 1

    return DeviceMesh(
        n_nodes=n_nodes,
        n_gpus_per_node=n_gpus_per_node,
        mapping=mapping,
        global_mesh_name=global_mesh_name,
        name=name,
    )


def device_mesh_name_from_mapping(global_mesh_name: str, mapping: np.ndarray):
    prefix = cluster_spec.node_name_prefix
    node_list = parse_nodelist(global_mesh_name, prefix)
    n_nodes = len(node_list)
    n_gpus_per_node = 8
    assert mapping.shape == (n_nodes, n_gpus_per_node)
    node_indices, gpu_ids = np.where(mapping == 1)

    if np.sum(mapping) < 8:
        node_name = node_list[node_indices[0]]
        gpu_ids = list(map(str, gpu_ids))
        return f"{node_name}:{','.join(gpu_ids)}"
    else:
        unique_node_indices = np.unique(node_indices)
        sub_node_list = [node_list[i] for i in unique_node_indices]
        node_name = nodelist_from_nodes(sub_node_list, prefix)
        return node_name


def find_parallel_strategies(
    device_mesh: DeviceMesh,
) -> List[ParallelismConfig]:
    n_gpus = np.sum(device_mesh.mapping)
    res = []
    for num_mp in [1, 2, 4, 8]:
        if n_gpus >= num_mp:
            assert n_gpus % num_mp == 0
            num_dp_pp = n_gpus // num_mp
            num_pp = 1
            while num_pp <= num_dp_pp:
                num_dp_mp = n_gpus // num_pp
                valid = (
                    num_dp_mp in [1, 2, 4, 8] or num_dp_mp % 8 == 0
                ) and num_dp_pp % num_pp == 0
                if valid:
                    res.append(
                        ParallelismConfig(num_pp, num_mp, num_dp_pp // num_pp)
                    )
                num_pp += 1
    return res


@dataclasses.dataclass
class RPCAllocation:
    rpc: Union[MFCDef, str]
    device_mesh: DeviceMesh
    parallel: ParallelismConfig

    def __post_init__(self):
        world_size = (
            self.parallel.model_parallel_size
            * self.parallel.pipeline_parallel_size
            * self.parallel.data_parallel_size
        )
        assert world_size == self.device_mesh.mapping.sum(), (
            "World size of ParallelismConfig does not match number of GPUs in device mesh"
            f"world_size {world_size} != n GPUs {self.device_mesh.mapping.sum()}"
        )

    def to_dict(self):
        return dict(
            rpc=self.rpc.name,
            device_mesh=self.device_mesh.to_dict(),
            parallel=dataclasses.asdict(self.parallel),
        )

    @staticmethod
    def from_dict(d):
        return RPCAllocation(
            rpc=d["rpc"],
            device_mesh=DeviceMesh.from_dict(d["device_mesh"]),
            parallel=ParallelismConfig(**d["parallel"]),
        )


@dataclasses.dataclass
class AllocationConfig:
    """Allocation configuration for one RPC, used for manual allocation.

    :param parallel: Parallelism strategy configuration.
    :type parallel: ParallelismConfig
    :param device_mesh: String representation for device mesh.
        If it is composed of multiple nodes, it should be in the form of
        slurm nodelist, e.g., node[01-02] or node01,node02.
        If it is a slice on a single node,
        we restrict it occupies 1, 2, 4, or 8 contiguous GPUs in the node.
        In this case, the string representation is similar to the MPI hostfile, e.g.,
        "node01:0-3" for the first 4 GPUs on node01.
    :type device_mesh: str
    """

    parallel: ParallelismConfig = dataclasses.field(
        default_factory=ParallelismConfig
    )
    device_mesh: Optional[str] = None



from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union
import collections
import dataclasses
import functools
import inspect
import json
import os
import time

import numpy as np
import torch.utils.data
import transformers

from realhf.api.core import model_api, system_api
from realhf.base import datapack, logging, namedarray
from realhf.base.cluster import spec as cluster_spec

logger = logging.getLogger("api.data")


@dataclasses.dataclass
class DataBatchMeta:
    dp_rank: int
    keys: List[str]
    seqlens: List[int]
    epoch: int
    is_final_batch: bool
    hash_vals: List[int]


@dataclasses.dataclass
class DatasetUtility:
    seed: int
    ddp_rank: int
    world_size: int
    tokenizer: transformers.PreTrainedTokenizerFast

    def __post_init__(self):
        if self.tokenizer.pad_token_id is None:
            self.tokenizer.pad_token_id = self.tokenizer.eos_token_id
            if self.tokenizer.eos_token_id is None:
                raise ValueError("eos_token_id of tokenizer must be defined.")


def unpack_data_batch(x: DataBatchMeta) -> List[DataBatchMeta]:
    return [
        DataBatchMeta(x.dp_rank, x.keys, [seqlen], x.epoch, False, [hash_val])
        for seqlen, hash_val in zip(x.seqlens, x.hash_vals)
    ]


def get_shuffle_indices(seed: int, size: int):
    """Generate shuffled indices given seed and (dataset) size."""
    np_rng = np.random.RandomState(seed=seed)
    dtype_ = np.uint32
    if size >= (np.iinfo(np.uint32).max - 1):
        dtype_ = np.int64
    shuffle_idx = np.arange(start=0, stop=size, step=1, dtype=dtype_)
    np_rng.shuffle(shuffle_idx)
    return shuffle_idx


def load_shuffle_split_dataset(
    util: DatasetUtility,
    dataset_path: str,
    dataset_builder: Optional[Callable[[], List[Dict[str, str]]]] = None,
):
    if dataset_path is not None:
        if dataset_path.endswith(".jsonl"):
            with open(dataset_path, "r") as f:
                data = [json.loads(ff) for ff in f]
        elif dataset_path.endswith(".json"):
            with open(dataset_path, "r") as f:
                data = json.load(f)
        else:
            raise NotImplementedError(
                f"Unkown dataset extension: {dataset_path}"
            )
    else:
        assert dataset_builder is not None
        data = dataset_builder()

    datasize_per_rank = len(data) // util.world_size
    shuffle_indices = get_shuffle_indices(
        util.seed, datasize_per_rank * util.world_size
    )
    subset_indices = shuffle_indices[
        util.ddp_rank
        * datasize_per_rank : (util.ddp_rank + 1)
        * datasize_per_rank
    ]
    data: List[Dict[str, str]] = [data[i] for i in subset_indices]

    return data


ALL_DATASET_CLASSES = {}


def register_dataset(name, dataset_cls):
    assert name not in ALL_DATASET_CLASSES
    assert "/" not in name
    # call_arg_names = list(inspect.signature(dataset_cls).parameters.keys())
    # if 'util' not in call_arg_names:
    #     raise KeyError(
    #         f"'util' must be one of the arguments in __init__, which is an instance of DatasetUtility. "
    #         f"Existing arguments: {call_arg_names}.")
    ALL_DATASET_CLASSES[name] = dataset_cls


def make_dataset(
    cfg: Union[str, system_api.Dataset],
    seed: int,
    ddp_rank: int,
    world_size: int,
    tokenizer_or_tokenizer_name: Union[
        transformers.PreTrainedTokenizerFast, str
    ],
    experiment_name: str,
    trial_name: str,
    cache_root: Optional[str] = None,
) -> torch.utils.data.Dataset:
    if isinstance(cfg, str):
        cfg = system_api.Dataset(type_=cfg)

    if isinstance(tokenizer_or_tokenizer_name, str):
        tokenizer = model_api.load_hf_tokenizer(tokenizer_or_tokenizer_name)
    elif tokenizer_or_tokenizer_name is None:
        raise RuntimeError("tokenizer_or_tokenizer_name cannot be None.")
    else:
        tokenizer = tokenizer_or_tokenizer_name
    util = DatasetUtility(
        seed,
        ddp_rank,
        world_size,
        tokenizer,
    )

    if cache_root is None:
        dataset_cls = ALL_DATASET_CLASSES[cfg.type_]
        return dataset_cls(util=util, **cfg.args)

    # Create and check cache path.
    if not cache_root.startswith(
        cluster_spec.fileroot
    ) and not cache_root.startswith("/home"):
        raise ValueError(
            f"Data cache path {cache_root} should be /home or under {cluster_spec.fileroot}."
        )
    if "_" in experiment_name or "_" in trial_name:
        raise ValueError(f"Invalid experiment/trial name.")

    output_path = os.path.join(
        cache_root,
        experiment_name,
        trial_name,
        cfg.type_,
        f"seed{seed}",
        f"world_size{world_size}",
        f"rank{ddp_rank}",
    )
    os.makedirs(output_path, exist_ok=True)

    fname = "dataset.pt"
    cache_found = os.path.isfile(os.path.join(output_path, fname))

    tik = time.perf_counter()
    if not cache_found:
        logger.info(
            f"No data cache found for rank {ddp_rank}. Create it from scratch."
        )
        dataset = ALL_DATASET_CLASSES[cfg.type_](
            seed, ddp_rank, world_size, **cfg.args
        )
        torch.save(dataset, os.path.join(output_path, fname))
    else:
        logger.info(f"Rank {ddp_rank} find existing data cache, load it.")
        dataset = torch.load(os.path.join(output_path, fname))
    logger.info(
        f"Dataset creation/loading time: {time.perf_counter() - tik:.3f}s"
    )

    return dataset


ALL_DATALOADER_CLASSES = {}


def register_dataloader(name, dataloader_cls):
    assert name not in ALL_DATALOADER_CLASSES
    ALL_DATALOADER_CLASSES[name] = dataloader_cls


def make_dataloader(
    cfg: Union[str, system_api.DataLoader], dataset: torch.utils.data.Dataset
) -> torch.utils.data.DataLoader:
    if isinstance(cfg, str):
        cfg = system_api.DataLoader(type_=cfg)
    dataloader_cls = ALL_DATALOADER_CLASSES[cfg.type_]
    return dataloader_cls(dataset, **cfg.args)


def PackedDataLoader(dataset, *args, **kwargs):
    return torch.utils.data.DataLoader(
        dataset,
        *args,
        collate_fn=gather_sequences,
        # NOTE: This is *NOT* the actual batch size for training.
        # It is just a proper size to load data to workers.
        batch_size=512,
        shuffle=True,
        **kwargs,
    )


def PackedEvalDataLoader(dataset, *args, **kwargs):
    return torch.utils.data.DataLoader(
        dataset,
        *args,
        collate_fn=gather_sequences,
        shuffle=False,
        **kwargs,
    )


register_dataloader("packed", PackedDataLoader)
register_dataloader("packed_eval", PackedEvalDataLoader)


def split_sequences(
    src: namedarray.NamedArray,
    n_dp: int = None,
    return_sizes: bool = False,
    partitions: Optional[List[Tuple[int, int]]] = None,
    min_size: int = 1,
    return_partitions=False,
) -> List[namedarray.NamedArray]:
    # FIXME: remove cu_seqlens here
    if src.metadata.get("seqlens", None) is None:
        raise ValueError(
            "seqlens must be in the metadata of the input namedarray."
        )

    seqlens = src.metadata["seqlens"]

    if n_dp is None:
        partitions = [(i, i + 1) for i in range(len(seqlens))]
        n_dp = len(seqlens)

    if partitions is None:
        partitions = datapack.min_abs_diff_partition(seqlens, n_dp, min_size)

    batch_sizes = [end - start for start, end in partitions]

    # These are used by log probabilities, which are one-step shorter than packed inputed ids.
    # We use numpy/list for indexing to avoid host-device synchronization
    batch_seqlens = [sum(seqlens[start:end]) for start, end in partitions]
    offsets = [0] + np.cumsum(batch_seqlens).tolist()[:-1]
    short1batch_seqlens = [
        sum(seqlens[start:end]) - (end - start) for start, end in partitions
    ]
    short1offsets = [0] + np.cumsum(short1batch_seqlens).tolist()[:-1]

    splitted_data = [collections.defaultdict() for _ in range(n_dp)]
    for k, v in src.items():
        for i, sp in enumerate(splitted_data):
            # NOTE: This is a terrible implementation, because we must know how to split each tensor according to its semantics.
            # Different tensor has different shape and semantics,
            # e.g., packed_seq has shape [tot_seqlen] and should be splited according to cumulative lengths,
            # packed_logprobs has shape [tot_seqlen - bs] (each sequence is one-step shorter) and should be splitted
            # according to short-1 cumulative lengths, seq_no_eos_mask has shape [bs] and performs similar as input_lens, ...
            # so we must enumerate each possible key and deal with them separately, etc.
            if v is None:
                sp[k] = None
            elif k in ["prompt_cu_seqlens", "cu_seqlens"]:
                continue
            elif k in [
                "prompt_lens",
                "input_lens",
                "seq_no_eos_mask",
                "rewards",
                "reward_score",
                "group_factor",
                "prompt_lens",
                "pos_input_lens",
                "group_input_lens",
                "seqlogp",
            ]:
                start, end = partitions[i]
                sp[k] = v[start:end]
            elif k in [
                "packed_seq",
                "packed_logits_mask",
                "prompt_mask",
                "packed_input_ids",
                "values",
                "logits_mask",
                "packed_prompts",
            ]:
                sp[k] = v[offsets[i] : offsets[i] + batch_seqlens[i]]
            elif k in [
                "packed_logprobs",
                "packed_ref_logprobs",
                "old_logp",
                "ref_logp",
                "advantages",
                "ppo_loss_mask",
                "kl_rewards",
                "returns",
            ]:
                sp[k] = v[
                    short1offsets[i] : short1offsets[i] + short1batch_seqlens[i]
                ]
            elif not torch.is_tensor(src[k]):
                # for constant, preserve value for each splitted instance
                sp[k] = src[k]
            else:
                raise RuntimeError(
                    f"Unknown key {k} in packed data. We don't know how to split it. "
                    f"Check api/core/data_api.py for implemented keys."
                )

    if "cu_seqlens" in src.keys():
        for x, (start, end) in zip(splitted_data, partitions):
            slens = torch.tensor(
                seqlens[start:end],
                dtype=torch.int32,
                device=src["cu_seqlens"].device,
            )
            x["cu_seqlens"] = torch.nn.functional.pad(
                slens.cumsum(dim=0), (1, 0)
            ).int()

    if "prompt_cu_seqlens" in src.keys():
        raw_prompt_lens = (
            src["prompt_cu_seqlens"][1:] - src["prompt_cu_seqlens"][:-1]
        )
        all_prompt_lens: List[torch.IntTensor] = [
            raw_prompt_lens[start:end].int() for start, end in partitions
        ]
        for x, pslens in zip(splitted_data, all_prompt_lens):
            x["prompt_cu_seqlens"] = torch.nn.functional.pad(
                pslens.cumsum(dim=0), (1, 0)
            ).int()

    splitted_data = [namedarray.from_dict(dict(x)) for x in splitted_data]
    for x, (start, end) in zip(splitted_data, partitions):
        x.register_metadata(seqlens=seqlens[start:end])

    res = [splitted_data]
    if return_sizes:
        res.append(batch_sizes)
    if return_partitions:
        res.append(partitions)

    if not return_sizes and not return_partitions:
        return res[0]
    else:
        return res


def gather_sequences(src: List[namedarray.NamedArray]) -> namedarray.NamedArray:
    seqlens = []
    for x in src:
        if x.metadata.get("seqlens", None) is None:
            raise ValueError(
                "seqlens must be in the metadata of the input namedarray."
            )
        seqlens += x.metadata["seqlens"]

    res = namedarray.recursive_aggregate(src, lambda x: torch.cat(x, dim=0))

    if "cu_seqlens" in src[0]:
        slens = torch.cat(
            [x["cu_seqlens"][1:] - x["cu_seqlens"][:-1] for x in src], dim=0
        )
        res["cu_seqlens"] = torch.nn.functional.pad(
            slens.cumsum(dim=0), (1, 0)
        ).int()

    if "prompt_cu_seqlens" in src[0]:
        slens = torch.cat(
            [
                x["prompt_cu_seqlens"][1:] - x["prompt_cu_seqlens"][:-1]
                for x in src
            ],
            dim=0,
        )
        res["prompt_cu_seqlens"] = torch.nn.functional.pad(
            slens.cumsum(dim=0), (1, 0)
        ).int()

    res.register_metadata(seqlens=seqlens)
    return res


def get_shape_from_key_and_seqlen(k: str, seqlen: int, vocab_size: int):
    if k in [
        "input_lens",
        "prompt_lens",
        "seq_no_eos_mask",
        "rewards",
        "reward_score",
        "group_factor",
        "pos_input_lens",
    ]:
        shape = (1,)
    elif k in ["cu_seqlens", "prompt_cu_seqlens"]:
        shape = (2,)
    # FIXME: problem here if we use groups instead of pairs?
    elif k in ["seqlogp"]:
        shape = (1, 2)
    elif k in [
        "packed_seq",
        "prompt_mask",
        "packed_input_ids",
        "values",
        "packed_prompts",
    ]:
        shape = (seqlen,)
    elif k in [
        "packed_logprobs",
        "packed_ref_logprobs",
        "old_logp",
        "ref_logp",
        "advantages",
        "ppo_loss_mask",
        "kl_rewards",
        "returns",
    ]:
        shape = (seqlen - 1,)
    elif k in ["logits_mask", "packed_logits_mask"]:
        shape = (seqlen, vocab_size)
    else:
        raise NotImplementedError(f"Unknown key {k} in packed data.")
    return shape


def get_dtype_from_key(k: str):
    if k in [
        "seq_no_eos_mask",
        "ppo_loss_mask",
        "prompt_mask",
        "logits_mask",
        "packed_logits_mask",
    ]:
        dtype = torch.bool
    elif k in [
        "reward_score",
        "packed_ref_logprobs",
        "old_logp",
        "ref_logp",
        "advantages",
        "kl_rewards",
        "returns",
        "values",
    ]:
        dtype = torch.float16
    elif k in [
        "input_lens",
        "prompt_lens",
        "cu_seqlens",
        "prompt_cu_seqlens",
        "pos_input_lens",
    ]:
        dtype = torch.int32
    elif k in ["packed_seq", "packed_input_ids", "packed_prompts"]:
        dtype = torch.int64
    elif k in ["rewards", "packed_logprobs", "group_factor", "seqlogp"]:
        dtype = torch.float32
    else:
        raise NotImplementedError(f"Unknown key {k} in packed data.")
    return dtype



from contextlib import closing
import socket


def find_free_port():
    """From, stackoverflow Issue 1365265"""
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def gethostname():
    return socket.gethostname()


def gethostip():
    return socket.gethostbyname(socket.gethostname())



from collections import defaultdict
from typing import *
import contextlib
import dataclasses
import functools

from omegaconf import MISSING
import numpy as np

from realhf.api.core.config import Dataset
from realhf.api.core.dfg import (
    MFCDef,
    ModelInterfaceType,
    OffloadHook,
    SyncParamHook,
)
from realhf.api.core.system_api import *
from realhf.api.quickstart.device_mesh import (
    AllocationConfig,
    DeviceMesh,
    make_device_mesh_from_name,
    RPCAllocation,
)
from realhf.api.quickstart.model import ModelTrainEvalConfig
from realhf.experiments.common.check import *
from realhf.experiments.common.utils import *
from realhf.search_engine.search import search_rpc_allocations
import realhf.base.logging as logging

logger = logging.getLogger("CommonExperimentConfig", "colored")


@dataclasses.dataclass
class CommonExperimentConfig(Experiment):
    """The common config for quickstart experiments.

    All members can be changed by the user in the command line,
    e.g.,

    .. code-block:: shell

        $ python3 -m realhf.apps.quickstart sft trial_name=my_trial seed=42 ...

    Recover mode is one of the followings\:

    - ``auto``\: automatically recover the last failed run.

    - ``save``\: save recover states if any error occurs.

    - ``resume``\: resume from saved recover states and save states if fail again.

    - ``disabled``\: do nothing but raise error when error occurs.

    Allocation mode is one of the followings\:

    - ``manual``\: manually allocate resources with experiment configs.

    - ``search``\: allocate resources and configure parallel strategies with search.

    - ``heuristic``\: allocate resources and configure parallel strategies with heuristic strategy.

    - ``pipe_data``\: allocate all models on all cluster nodes available and configure parallel strategies with pipe+data parallelism.

    - ``pipe_model``: allocate all models on all cluster nodes available and configure parallel strategies with pipe+model parallelism.

    :param experiment_name: Name of the experiment.
    :type experiment_name: str
    :param trial_name: Name of the trial.
    :type trial_name: str
    :param mode: Experiment launching mode.
        Currently only "local" and "slurm" are supported.
        The "local" mode implies ``n_nodes=1``.
    :type mode: str
    :param debug: Whether to run in the debug mode.
        The non-debug mode will disable all assertions.
    :type debug: bool
    :param partition: The slurm partition to run the experiment.
    :type partition: str
    :param wandb_mode: The mode of wandb. Currently the wandb logging is not supported.
    :type wandb_mode: str
    :param image_name: The name of the docker image used by the controller.
        Only used in the slurm mode.
    :type image_name: str or None
    :param recover_mode: The recover mode.
    :type recover_mode: str
    :param recover_retries: The number of retries for recovery.
        Only effective when recover_mode is "auto".
    :type recover_retries: int
    :param ignore_worker_error: Whether to ignore errors raised by
        workers during runtime. Please do not set it to be True unless
        the user is sure that some error is ignorable.
        Only effective when recover_mode is "disabled".
    :type ignore_worker_error: bool
    :param allocation_mode: Mode of GPU parallel strategy allocation.
    :type allocation_mode: str
    :param allocation_use_cache: Whether to use cache in allocation search.
        Only effective when allocation_mode=="search"
        and cache is available in the log dir of current experiment
        name and trial.
    :type allocation_use_cache: bool
    :param n_nodes: Number of nodes to run the experiment.
        Only effective when mode=="slurm".
    :type n_nodes: int
    :param n_gpus_per_node: Number of GPUs per node.
        Only effective when mode=="slurm".
    :type n_gpus_per_node: int
    :param nodelist: Slurm nodelist.
        Only effective when mode=="slurm".
    :type nodelist: str or None
    :param seed: Random seed.
    :type seed: int
    """

    experiment_name: str = MISSING
    trial_name: str = MISSING
    mode: str = "slurm"
    debug: bool = True
    partition: str = "dev"
    wandb_mode: str = "disabled"
    image_name: Optional[str] = None
    recover_mode: str = "disabled"
    recover_retries: int = 1
    ignore_worker_error: bool = False
    allocation_mode: str = "pipe_model"
    allocation_use_cache: bool = False
    n_nodes: int = 1
    n_gpus_per_node: int = 8
    nodelist: Optional[str] = None
    seed: int = 1

    @property
    def models(self) -> Dict[str, ModelTrainEvalConfig]:
        raise NotImplementedError(
            f"models is not implemented in {self.__class__}"
        )

    @property
    def rpcs(self) -> Dict[str, MFCDef]:
        raise NotImplementedError(
            f"rpcs is not implemented in {self.__class__}"
        )

    @property
    def datasets(self) -> List[Dataset]:
        return []

    @property
    def eval_datasets(self) -> List[Dataset]:
        return None

    @property
    def eval_dataloader(self) -> DataLoader:
        return DataLoader("packed_eval", args=dict(batch_size=128))

    @property
    def tokenizer_name_or_path(self) -> str:
        raise NotImplementedError(
            f"tokenizer_name_or_path is not implemented in {self.__class__}"
        )

    @property
    def exp_ctrl(self) -> ExperimentSaveEvalControl:
        raise NotImplementedError(
            f"expr_ctrl is not implemented in {self.__class__}"
        )

    @property
    def max_prompt_len(self) -> int:
        return None

    @property
    def search_kwargs(self) -> Dict[str, Any]:
        return {}

    @property
    def allocations(self) -> Dict[str, AllocationConfig]:
        return {}

    @property
    def global_device_mesh(self) -> DeviceMesh:
        return DeviceMesh(
            n_nodes=self.n_nodes,
            n_gpus_per_node=self.n_gpus_per_node,
            mapping=np.ones(
                (self.n_nodes, self.n_gpus_per_node), dtype=np.int32
            ),
            global_mesh_name=self.nodelist,
            name=self.nodelist,
        )

    def _heuristic_rpc_allocation(self) -> List[RPCAllocation]:
        raise NotImplementedError(
            f"_heuristic_rpc_allocation is not implemented in {self.__class__}"
        )

    def _search(self):
        # called in both api.main and controller
        gradient_checkpointing = any(
            model.gradient_checkpointing for model in self.models.values()
        )
        rpc_allocs: List[RPCAllocation] = search_rpc_allocations(
            device_mesh=self.global_device_mesh,
            rpcs=list(self.rpcs.values()),
            gradient_checkpointing=gradient_checkpointing,
            use_cache=self.allocation_use_cache,
            **self.search_kwargs,
        )
        return rpc_allocs

    def scheduling_setup(self) -> ExperimentScheduling:
        return ExperimentScheduling(
            master_worker=TasksGroup(
                count=1,
                scheduling=Scheduling.master_worker_default(
                    cpu=4, mem=20000, nodelist=self.nodelist
                ),
            ),
            model_worker=TasksGroup(
                count=self.n_nodes * self.n_gpus_per_node,
                scheduling=Scheduling.model_worker_default(
                    cpu=4,
                    gpu=1,
                    gpu_type="tesla",
                    mem=100000,
                    nodelist=self.nodelist,
                ),
            ),
        )

    def initial_setup(self) -> ExperimentConfig:
        if self.allocation_mode == "manual" and self.nodelist is None:
            logger.warning(
                "Warning: Nodelist is not set in manual allocation mode, "
                "in this case you cannot specify device mesh for each model RPC. "
                "All model RPC will be allocated on GPUs automatically "
                f"allocated according to n_nodes {self.n_nodes} "
                f"and n_gpus_per_node {self.n_gpus_per_node}."
            )
        self.__check_legal_experiment()

        rpcs = self.rpcs
        model_worker = []

        if self.allocation_mode == "search":
            # assert self.mode == "slurm"
            # assumes gradient checkpointing for all training RPCs if one is enabled
            # for the simplicity of search configurations
            rpc_allocs = self._search()
            for rpc_alloc in rpc_allocs:
                assert isinstance(rpc_alloc.rpc, str)
                for rpc in rpcs.values():
                    if rpc.name == rpc_alloc.rpc:
                        rpc_alloc.rpc = rpc
                        break
                else:
                    raise ValueError(f"RPC {rpc_alloc.rpc} not found in rpcs.")
        elif (
            self.allocation_mode == "pipe_data"
            or self.allocation_mode == "pipe_model"
        ):
            rpc_allocs: List[RPCAllocation] = [
                RPCAllocation(
                    rpc=rpc,
                    device_mesh=self.global_device_mesh,
                    parallel=ParallelismConfig(
                        data_parallel_size=(
                            self.n_gpus_per_node
                            if self.allocation_mode == "pipe_data"
                            else 1
                        ),
                        pipeline_parallel_size=self.n_nodes,
                        model_parallel_size=(
                            self.n_gpus_per_node
                            if self.allocation_mode == "pipe_model"
                            else 1
                        ),
                        use_sequence_parallel=(
                            rpc.interface_type == ModelInterfaceType.TRAIN_STEP
                            and self.allocation_mode == "pipe_model"
                        ),
                    ),
                )
                for rpc in self.rpcs.values()
            ]
        elif self.allocation_mode == "manual":
            rpc_allocs: List[RPCAllocation] = [
                RPCAllocation(
                    rpc=rpc,
                    device_mesh=(
                        make_device_mesh_from_name(
                            self.nodelist,
                            self.allocations[rpc_type].device_mesh,
                        )
                        if self.allocations[rpc_type].device_mesh is not None
                        else self.global_device_mesh
                    ),
                    parallel=self.allocations[rpc_type].parallel,
                )
                for rpc_type, rpc in self.rpcs.items()
            ]
        elif self.allocation_mode == "heuristic":
            rpc_allocs: List[RPCAllocation] = self._heuristic_rpc_allocation()
        else:
            raise NotImplementedError()

        shard_counter = defaultdict(lambda: 0)
        resolve_rpc_hooks(
            rpc_allocs
        )  # inplace modify MFCDefs in rpc allocations
        import pprint

        pprint.pprint(rpc_allocs)

        model_name_to_rpc_allocs: Dict[ModelName, List[RPCAllocation]] = (
            defaultdict(list)
        )
        for rpc_alloc in rpc_allocs:
            model_name_to_rpc_allocs[rpc_alloc.rpc.model_name].append(rpc_alloc)

        for i, j in itertools.product(
            range(self.n_nodes), range(self.n_gpus_per_node)
        ):
            mw = ModelWorker(
                seed=self.seed,
                shards=[],
                datasets=self.datasets,
                cuda_cache_cleanliness=False,
                cuda_cache_clear_freq=10,
                tokenizer_name_or_path=self.tokenizer_name_or_path,
            )
            # print(f"Setting up ModelWorker ({i},{j})")

            for (
                model_name,
                model_rpc_allocs,
            ) in model_name_to_rpc_allocs.items():
                rpcs = [rpc_alloc.rpc for rpc_alloc in model_rpc_allocs]
                rpc_alloc = model_rpc_allocs[0]
                model_cfg = self.models[model_name.role]
                model = make_model_config(model_cfg)
                mapping = rpc_alloc.device_mesh.mapping
                gradient_checkpointing = (
                    model_cfg.gradient_checkpointing
                    and any(
                        rpc.interface_type == ModelInterfaceType.TRAIN_STEP
                        for rpc in rpcs
                    )
                )

                topo = get_topo(
                    rpc_alloc.parallel,
                    gradient_checkpointing=gradient_checkpointing,
                    max_prompt_len=(
                        self.max_prompt_len
                        if any(
                            rpc.interface_type == ModelInterfaceType.GENERATE
                            for rpc in rpcs
                        )
                        else None
                    ),
                )

                if any(
                    rpc.interface_type == ModelInterfaceType.TRAIN_STEP
                    for rpc in rpcs
                ):
                    backend = make_train_backend_config(
                        model_cfg, rpc_alloc.parallel
                    )
                else:
                    backend = make_inf_backend_config(
                        model_cfg, rpc_alloc.parallel
                    )

                # print(f"model name {model_name}, device mesh name {rpc_alloc.device_mesh.name},"
                #       f"mapping {mapping}")
                if mapping[i, j]:
                    shard_idx = shard_counter[model_name]
                    # print(f"Setting up Shard {shard_idx}, "
                    #       f"(dp, pp, mp) = ({topo.get_coord(shard_idx).data}, "
                    #       f"{topo.get_coord(shard_idx).pipe}, {topo.get_coord(shard_idx).model})")
                    mw.shards.append(
                        StandaloneModelShard(
                            id=ModelShardID(
                                model_name=model_name,
                                topo=topo,
                                dp_rank=topo.get_coord(shard_idx).data,
                                pp_rank=topo.get_coord(shard_idx).pipe,
                                mp_rank=topo.get_coord(shard_idx).model,
                            ),
                            model=model,
                            backend=backend,
                            eval_datasets=self.eval_datasets,
                            eval_dataloader=self.eval_dataloader,
                        )
                    )
                    shard_counter[model_name] += 1
            model_worker.append(mw)

        return ExperimentConfig(
            exp_ctrl=self.exp_ctrl,
            model_rpcs=[rpc_alloc.rpc for rpc_alloc in rpc_allocs],
            model_worker=model_worker,
        )

    def __check_legal_experiment(self):
        if self.n_nodes > 1 and self.mode == "local":
            raise ValueError(
                "Cannot run multi-node experiment in local mode, "
                "please setup slurm for distributed runs."
            )

        if self.n_gpus_per_node != 8 and self.allocation_mode in [
            "search",
            "heuristic",
        ]:
            raise ValueError(
                f"Cannot run search or heuristic allocation with "
                f"n_gpus_per_node {self.n_gpus_per_node}, "
                "please set n_gpus_per_node to 8."
            )

        for rpc_name, rpc in self.rpcs.items():
            if not check_is_realhf_native_model_interface(
                rpc.interface_impl.type_
            ) and self.allocation_mode in ["search", "heuristic"]:
                raise ValueError(
                    f"RPC {rpc.name} interface is not a realhf native implementation. "
                    f"Search and heuristic allocation mode are not available."
                )
            if (
                self.allocation_mode == "manual"
                and rpc_name not in self.allocations
            ):
                if rpc_name not in self.allocations:
                    raise ValueError(
                        f"RPC {rpc_name} is not in allocations, please implement "
                        f"`allocations()` method in your config class to enable "
                        f"manual allocation."
                    )

            if rpc.model_name.role not in self.models.keys():
                raise ValueError(
                    f"RPC {rpc.name} model name {rpc.model_name.role} is not in models."
                )



from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.utils.checkpoint

from realhf.impl.model.parallelism.model_parallel.modules import (
    RowParallelLinear,
)
from realhf.impl.model.utils.functional import (
    apply_rotary_varlen,
    compute_varlen_position_indices,
    torch_attn_func,
)
import realhf.base.logging as logging

from .mlp import LayerNormQKVLinear
from .rotary import RotaryEmbedding

try:
    from flash_attn import (
        flash_attn_func,
        flash_attn_varlen_func,
        flash_attn_with_kvcache,
    )
except ModuleNotFoundError:
    pass
import realhf.base.logging as logging

logger = logging.getLogger("Attention")


class CausalSelfAttentionLayer(nn.Module):

    def __init__(
        self,
        hidden_dim: int,
        n_kv_heads: int,
        head_dim: int,
        resid_pdrop: float,
        attn_pdrop: float,
        layer_index: int,
        layer_norm_epsilon: float,
        # gpt2 does not scale attn by inverse layer idx, in contrast to starcoder
        scale_attn_by_inverse_layer_idx: bool,
        # llama does not require attention bias
        use_attention_bias: bool,
        # layer norm type is special for llama
        layer_norm_type: Optional[str] = None,
        # rotary embedding
        apply_rotary: bool = False,
        rotary_base: float = 10000.0,
        rotary_interleaved: bool = False,  # False for LLaMA, GPT-neoX; True for GPT-J
        rotary_scaling: Optional[float] = None,
        rotary_scaling_type: Optional[str] = None,
        # parallel settings
        model_parallel: bool = False,
        gradient_accumulation_fusion: bool = False,
        # device and dtype
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super().__init__()
        if dtype is None:
            dtype = torch.float16
        assert hidden_dim % head_dim == 0
        n_q_heads = hidden_dim // head_dim
        self.c_attn = LayerNormQKVLinear(
            input_dim=hidden_dim,
            head_dim=head_dim,
            n_q_heads=n_q_heads,
            n_kv_heads=n_kv_heads,
            model_parallel=model_parallel,
            gradient_accumulation_fusion=gradient_accumulation_fusion,
            layer_norm_epsilon=layer_norm_epsilon,
            layer_norm_type=layer_norm_type,
            use_attention_bias=use_attention_bias,
            dtype=dtype,
            device=device,
            layer_index=layer_index,
        )

        if model_parallel:
            self.c_proj = RowParallelLinear(
                hidden_dim,
                hidden_dim,
                bias=use_attention_bias,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                dtype=dtype,
                device=device,
            )
        else:
            self.c_proj = nn.Linear(
                hidden_dim,
                hidden_dim,
                bias=use_attention_bias,
                dtype=dtype,
                device=device,
            )

        self.resid_dropout = nn.Dropout(resid_pdrop)

        self.attn_pdrop = attn_pdrop

        self.applied_attn_pdrop = attn_pdrop

        self.apply_rotary = apply_rotary
        self.rotary_interleaved = rotary_interleaved
        if self.apply_rotary:
            # Will layzily update the cache sequence length of cache.,
            # so we don't need to pass in max_positions.
            self.rotary_emb = RotaryEmbedding(
                head_dim,
                base=rotary_base,
                scale_factor=rotary_scaling,
                scale_type=rotary_scaling_type,
                interleaved=rotary_interleaved,
                device=device,
            )

        # constant
        self.h = hidden_dim
        self.nq = n_q_heads
        self.nkv = n_kv_heads
        if self.nq % self.nkv != 0:
            raise ValueError(
                f"n_kv_heads ({self.nkv}) must divide n_q_heads ({self.nq})."
            )
        self.d = head_dim

        self.layer_index = layer_index

        self.scale_attn_by_inverse_layer_idx = scale_attn_by_inverse_layer_idx

    def train(self, mode: bool):
        if not mode:
            self.applied_attn_pdrop = 0.0
        else:
            self.applied_attn_pdrop = self.attn_pdrop
        super().train(mode)
        return self

    def forward(
        self,
        hidden_states: torch.Tensor,
        cu_seqlens: Optional[torch.Tensor] = None,
        k_cache: Optional[torch.Tensor] = None,
        v_cache: Optional[torch.Tensor] = None,
        cache_seqlens: Optional[Union[int, torch.Tensor]] = None,
        max_seqlen: Optional[int] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        # input shape: [bs, seq, hidden_dim]

        # NOTE: we must ensure the passed-in argument is an interger
        # if we convert the argument to implicitly when calling rotary embedding or flash-attn,
        # aten::item will be called, which will cause a device-host sync and slow down performance.
        assert max_seqlen is None or isinstance(max_seqlen, int), type(
            max_seqlen
        )
        assert cu_seqlens is None or cu_seqlens.dtype == torch.int32

        # default upcast, scale
        if self.scale_attn_by_inverse_layer_idx:
            unscale = self.layer_index + 1
            scale_factor = unscale**-1
        else:
            unscale = 1.0
            scale_factor = 1
        scale_factor /= self.d**0.5

        q, k, v = self.c_attn(hidden_states)

        if self.apply_rotary and (k_cache is None or str(q.device) == "cpu"):
            # otherwise, we input rotary cos/sin directly into flash_attn_with_kvcache
            rotary_cache_len = max_seqlen
            if k_cache is not None and str(q.device) == "cpu":
                rotary_cache_len = k_cache.shape[1]
            self.rotary_emb._update_cos_sin_cache(
                rotary_cache_len, q.device, q.dtype
            )
            rotary_indices = compute_varlen_position_indices(
                q.shape[0], cu_seqlens
            )
            qk = apply_rotary_varlen(
                torch.cat([q, k], dim=-2),
                cos=self.rotary_emb._cos_cached,
                sin=self.rotary_emb._sin_cached,
                cu_seqlens=cu_seqlens,
                interleaved=self.rotary_emb.interleaved,
                rotary_indices=rotary_indices,
                seqlen_offsets=cache_seqlens,
            )
            # HACK: RotaryEmbedding used flash-attention's triton kernel internally, but it will
            # cause an illegal memory access error when batch size is large. Use pytorch implementation instead.
            # qk = self.rotary_emb(
            #     torch.cat([q, k], dim=-2),
            #     cu_seqlens=cu_seqlens,
            #     max_seqlen=max_seqlen,
            # )
            q, k = qk.split((q.shape[-2], k.shape[-2]), dim=-2)
        elif self.apply_rotary:
            self.rotary_emb._update_cos_sin_cache(
                k_cache.shape[1], device=q.device, dtype=q.dtype
            )
            # Rotary cos/sin will be automatically offset by cache_seqlens in flash_attn.
            rotary_cos, rotary_sin = (
                self.rotary_emb._cos_cached,
                self.rotary_emb._sin_cached,
            )
        else:
            rotary_cos = rotary_sin = None

        if str(q.device) == "cpu":
            cu_seqlens_k = cu_seqlens
            max_seqlen_k = max_seqlen
            if k_cache is not None:
                new_k, new_v = [], []
                for i, cache_len in enumerate(cache_seqlens):
                    k_cache[i, cache_len] = k[cu_seqlens[i] : cu_seqlens[i + 1]]
                    new_k.append(k_cache[i, : cache_len + 1])
                    v_cache[i, cache_len] = v[cu_seqlens[i] : cu_seqlens[i + 1]]
                    new_v.append(v_cache[i, : cache_len + 1])
                k = torch.cat(new_k, dim=0)
                v = torch.cat(new_v, dim=0)
                cu_seqlens_k = torch.nn.functional.pad(
                    (cache_seqlens + 1).cumsum(0), (1, 0)
                )
                max_seqlen_k = max(cache_seqlens) + 1
            # Use vanilla pytorch attention, for debugging.
            hidden_states = torch_attn_func(
                q,
                k,
                v,
                causal=True,
                cu_seqlens_q=cu_seqlens,
                max_seqlen_q=max_seqlen,
                cu_seqlens_k=cu_seqlens_k,
                max_seqlen_k=max_seqlen_k,
                dropout_p=self.applied_attn_pdrop,
                softmax_scale=scale_factor,
                upcast_unscale=unscale,
            )
        elif k_cache is not None:
            # k_cache/v_cache shape: [bs, max_seq, n_kv_heads, head_dim]
            if cache_seqlens is None:
                raise RuntimeError(
                    "cache_seqlens must be provided if kv_cache is not None."
                )
            q = q.unsqueeze(1)
            k = k.unsqueeze(1)
            v = v.unsqueeze(1)
            # k_cache and v_cache will be modified in-place.
            hidden_states = flash_attn_with_kvcache(
                q,
                k_cache,
                v_cache,
                k=k,
                v=v,
                cache_seqlens=cache_seqlens,
                softmax_scale=scale_factor,
                causal=False,  # True or False doesn't matter because seqlen=1
                rotary_cos=rotary_cos,
                rotary_sin=rotary_sin,
                rotary_interleaved=self.rotary_interleaved,
            )
            hidden_states = hidden_states.squeeze(1)
        elif cu_seqlens is not None:
            assert max_seqlen is not None
            assert len(q.shape) == 3
            hidden_states = flash_attn_varlen_func(
                q,
                k,
                v,
                cu_seqlens,
                cu_seqlens,
                max_seqlen,
                max_seqlen,
                dropout_p=self.applied_attn_pdrop,
                softmax_scale=scale_factor,
                causal=True,
            )
        else:
            raise NotImplementedError(
                "Don't know which attention implementation to use."
            )
        hidden_states = self.c_proj(hidden_states.flatten(start_dim=-2))
        hidden_states = self.resid_dropout(hidden_states)
        return hidden_states, k, v



import dataclasses

from realhf.api.core.dfg import (
    MFCDef,
    ModelInterface,
    ModelInterfaceType,
    OffloadHook,
)
from realhf.api.core.system_api import *
from realhf.api.quickstart.dataset import PairedComparisonDatasetConfig
from realhf.api.quickstart.device_mesh import AllocationConfig
from realhf.api.quickstart.entrypoint import register_quickstart_exp
from realhf.api.quickstart.model import (
    get_real_model_config,
    ModelTrainEvalConfig,
)
from realhf.base.topology import PipeModelDataParallelTopology
from realhf.experiments.common.common import CommonExperimentConfig
import realhf.base.logging as logging

logger = logging.getLogger("DPO Experiment")


@dataclasses.dataclass
class DPOConfig(CommonExperimentConfig):
    """Direct Preference Optimization (DPO) experiment configuration.

    It is a subclass of :class:`CommonExperimentConfig`,
    so all CLI options in the base class are available.

    We don't implement runtime evaluation for DPO.

    :param total_train_epochs: Total number of training epochs
        (i.e., the number of times the training dataset is iterated).
    :type total_train_epochs: int
    :param save_freq_steps: Save the model every this number of steps.
        "step" is an optimizer step or a single update of model parameters.
        If None, the model will not be saved during training.
        The directory to save the model will be automatically resolved
        and prompted in the terminal when the experiment starts.
    :type save_freq_steps: Optional[int]
    :param is_sft_lora: Whether LoRA was used for SFT.
        If so, the saved SFT model should only contain LoRA parameters.
        Since LoRA is currently not supported for SFT,
        this option is not used for now.
    :type is_sft_lora: bool
    :param sft_lora_path: Path to the LoRA model for SFT.
        Since LoRA is currently not supported for SFT,
        this option is not used for now.
    :param actor: Runtime configuration of the primary LLM.
    :type actor: ModelTrainEvalConfig
    :param ref: Runtime configuration of the reference LLM.
        This model is only used for inference to provide KL regularization.
        In ReaL, this model is automatically offloaded to CPU.
        As a result, training DPO is as efficient as training a single LLM.
    :type ref: ModelTrainEvalConfig
    :param actor_train: Device allocation and parallelism configuration for
        running training on the primary LLM.
    :type actor_train: AllocationConfig
    :param ref_inf: Device allocation and parallelism configuration for
        running inference on the reference LLM.
        This can be different from the training allocation.
        A large data parallel degree with additional pipelining
        can be faster for inference.
    :type ref_inf: AllocationConfig
    :param dataset: Dataset configuration. The same for reward modeling.
    :type dataset: PairedComparisonDatasetConfig
    :param beta: KL regularization coefficient.
    :type beta: float
    """

    total_train_epochs: int = 1
    save_freq_steps: Optional[int] = 20
    is_sft_lora: bool = False
    sft_lora_path: Optional[str] = None

    actor: ModelTrainEvalConfig = dataclasses.field(
        default_factory=ModelTrainEvalConfig
    )
    ref: ModelTrainEvalConfig = dataclasses.field(
        default_factory=ModelTrainEvalConfig
    )

    actor_train: AllocationConfig = dataclasses.field(
        default_factory=AllocationConfig
    )
    ref_inf: AllocationConfig = dataclasses.field(
        default_factory=AllocationConfig
    )

    dataset: PairedComparisonDatasetConfig = dataclasses.field(
        default_factory=PairedComparisonDatasetConfig
    )
    beta: float = 0.1

    def __post_init__(self):
        assert (
            not self.is_sft_lora and self.sft_lora_path is None
        ), "LoRA is not supported for now."

    @property
    def models(self):
        return {
            "actor": self.actor,
            "ref": self.ref,
        }

    @property
    def rpcs(self):
        interface = ModelInterface(
            "dpo", args=dict(beta=self.beta, enable_save=True)
        )
        ref_interface = ModelInterface(
            "dpo", args=dict(beta=self.beta, enable_save=False)
        )
        ref_inf = MFCDef(
            model_name=ModelName("ref", 0),
            interface_type=ModelInterfaceType.INFERENCE,
            interface_impl=ref_interface,
            model_type=self.ref.type,
            model_path=self.ref.path,
            input_data=[
                "packed_input_ids",
                "pos_input_lens",
                "prompt_lens",
            ],
            output_data=["seqlogp"],
            min_n_seqs=self.dataset.train_bs_n_seqs,
            max_n_seqs=self.dataset.train_bs_n_seqs,
        )
        dpo = MFCDef(
            model_name=ModelName("actor", 0),
            interface_type=ModelInterfaceType.TRAIN_STEP,
            interface_impl=interface,
            model_type=self.actor.type,
            model_path=self.actor.path,
            input_data=[
                "packed_input_ids",
                "pos_input_lens",
                "seqlogp",
                "prompt_lens",
            ],
            log_return_value=True,
            min_n_seqs=self.dataset.train_bs_n_seqs,
            max_n_seqs=self.dataset.train_bs_n_seqs,
        )
        return {
            "dpo": dpo,
            "ref_inf": ref_inf,
        }

    @property
    def allocations(self):
        return {
            "dpo": self.actor_train,
            "ref_inf": self.ref_inf,
        }

    @property
    def datasets(self):
        return [
            Dataset(
                "rw_pair",
                args=dict(
                    max_length=self.dataset.max_seqlen,
                    max_pairs_per_prompt=self.dataset.max_pairs_per_prompt,
                    dataset_path=self.dataset.train_path,
                ),
            )
        ]

    @property
    def tokenizer_name_or_path(self):
        return self.actor.path

    @property
    def exp_ctrl(self):
        return ExperimentSaveEvalControl(
            total_train_epochs=self.total_train_epochs,
            save_frequency_steps=self.save_freq_steps,
        )


register_quickstart_exp("dpo", DPOConfig)



from typing import Dict
import dataclasses

import torch
import torch.distributed as dist
import torch.utils.data
import tqdm

from realhf.base import constants
from realhf.base.namedarray import from_dict, NamedArray, recursive_apply
from realhf.impl.model.nn.real_llm_api import ReaLModel
from realhf.impl.model.utils.functional import gather_packed_shifted_log_probs
import realhf.api.core.model_api as model_api
import realhf.base.logging as logging
import realhf.impl.model.utils.dpo_functional as dpo_functional

logger = logging.getLogger("Packed DPO Interface")


def _dpo_loss_from_model_outputs(
    logits: torch.FloatTensor,
    packed_input_ids: torch.LongTensor,
    cu_seqlens: torch.IntTensor,
    prompt_lens: torch.IntTensor,
    seqlogp: torch.FloatTensor,
    beta: float,
    **kwargs,
):
    input_lens = cu_seqlens[1:] - cu_seqlens[:-1]
    logprobs = gather_packed_shifted_log_probs(
        logits, cu_seqlens, packed_input_ids
    ).float()

    assert (prompt_lens > 0).all(), prompt_lens
    logprob_sum = []
    offset = 0
    for i in range(prompt_lens.shape[0]):
        logprob_sum.append(
            logprobs[
                offset + prompt_lens[i] - 1 : offset + input_lens[2 * i] - 1
            ].sum()
        )
        offset += input_lens[2 * i] - 1
        logprob_sum.append(
            logprobs[
                offset + prompt_lens[i] - 1 : offset + input_lens[2 * i + 1] - 1
            ].sum()
        )
        offset += input_lens[2 * i + 1] - 1
    assert offset == sum(input_lens) - input_lens.shape[0], (
        offset,
        sum(input_lens),
        input_lens.shape,
    )

    pi_seqlogp = torch.stack(logprob_sum)

    loss, pos_score, neg_score, kl = dpo_functional.dpo_loss(
        pi_logps=pi_seqlogp,
        ref_logps=seqlogp.view(-1),
        beta=beta,
    )

    # Logging.
    logging_loss = (loss * prompt_lens.shape[0]).detach()
    n_seqs = torch.tensor(
        [prompt_lens.shape[0]], dtype=torch.float32, device=loss.device
    )
    dist.all_reduce(
        n_seqs, op=dist.ReduceOp.SUM, group=constants.data_parallel_group()
    )
    dist.all_reduce(
        pos_score, op=dist.ReduceOp.SUM, group=constants.data_parallel_group()
    )
    dist.all_reduce(
        neg_score, op=dist.ReduceOp.SUM, group=constants.data_parallel_group()
    )
    dist.all_reduce(
        logging_loss,
        op=dist.ReduceOp.SUM,
        group=constants.data_parallel_group(),
    )
    dist.all_reduce(
        kl, op=dist.ReduceOp.SUM, group=constants.data_parallel_group()
    )

    return loss, dict(
        loss=logging_loss,
        pos_score=pos_score,
        neg_score=neg_score,
        n_seqs=n_seqs,
        kl=kl,
    )


@dataclasses.dataclass
class DPOInterface(model_api.ModelInterface):
    beta: float
    enable_save: bool = True

    @torch.no_grad()
    def inference(self, model: model_api.Model, data: NamedArray) -> NamedArray:
        module = model.module
        module.eval()
        data = recursive_apply(data, lambda x: x.to(model.device))

        n_pairs = data["pos_input_lens"].shape[0]
        pair_lens = torch.tensor(
            data.metadata["seqlens"], dtype=torch.int32, device=model.device
        )
        input_lens: torch.IntTensor = torch.stack(
            [data["pos_input_lens"], pair_lens - data["pos_input_lens"]], 1
        ).view(-1)
        prompt_lens: torch.IntTensor = data["prompt_lens"]
        cu_seqlens = torch.cat(
            [input_lens.new_zeros(1), input_lens.cumsum(0)]
        ).int()
        input_lens = cu_seqlens[1:] - cu_seqlens[:-1]
        max_seqlen = int(max(input_lens))

        seqlens_cpu = input_lens.cpu().numpy().tolist()

        logits = module.forward(
            seqlens_cpu=seqlens_cpu,
            packed_input_ids=data["packed_input_ids"],
            cu_seqlens=cu_seqlens,
        )
        if logits is None:
            return None

        logprobs = gather_packed_shifted_log_probs(
            logits, cu_seqlens, data["packed_input_ids"]
        ).float()

        assert (prompt_lens > 0).all(), prompt_lens
        logprob_sum = []
        offset = 0
        for i in range(prompt_lens.shape[0]):
            logprob_sum.append(
                logprobs[
                    offset + prompt_lens[i] - 1 : offset + input_lens[2 * i] - 1
                ].sum()
            )
            offset += input_lens[2 * i] - 1
            logprob_sum.append(
                logprobs[
                    offset
                    + prompt_lens[i]
                    - 1 : offset
                    + input_lens[2 * i + 1]
                    - 1
                ].sum()
            )
            offset += input_lens[2 * i + 1] - 1
        assert offset == sum(input_lens) - input_lens.shape[0], (
            offset,
            sum(input_lens),
            input_lens.shape,
        )

        x = from_dict(dict(seqlogp=torch.stack(logprob_sum).view(n_pairs, -1)))
        x.register_metadata(**data.metadata)
        return x

    def train_step(self, model: model_api.Model, data: NamedArray) -> Dict:
        data = recursive_apply(data, lambda x: x.to(model.device))

        packed_input_ids: torch.Tensor = data["packed_input_ids"]
        pair_lens = torch.tensor(
            data.metadata["seqlens"], dtype=torch.int32, device=model.device
        )
        neg_input_lens = pair_lens - data["pos_input_lens"]
        input_lens: torch.Tensor = torch.stack(
            [data["pos_input_lens"], neg_input_lens], 1
        ).view(-1)
        prompt_lens: torch.IntTensor = data["prompt_lens"]
        cu_seqlens = torch.cat(
            [input_lens.new_zeros(1), input_lens.cumsum(0)], 0
        ).int()
        max_seqlen = int(max(cu_seqlens[1:] - cu_seqlens[:-1]))

        module = model.module
        module.eval()

        loss_fn_kwargs = dict(
            input_lens=pair_lens,
            prompt_lens=prompt_lens,
            seqlogp=data["seqlogp"],
            beta=self.beta,
        )
        stats = module.train_batch(
            seqlens_cpu=data.metadata["seqlens"],
            packed_input_ids=packed_input_ids,
            cu_seqlens=cu_seqlens,
            input_lens_for_partition=pair_lens,
            version_steps=model.version.global_step,
            loss_fn=_dpo_loss_from_model_outputs,
            **loss_fn_kwargs,
        )

        cur_epoch = model.version.epoch
        model.inc_version()

        res = {}
        if stats:
            res = dict(
                loss=float(stats["loss"]) / int(stats["n_seqs"]),
                pos_score=float(stats["pos_score"]) / int(stats["n_seqs"]),
                neg_score=float(stats["neg_score"]) / int(stats["n_seqs"]),
                kl=float(stats["kl"]) / int(stats["n_seqs"]),
                n_seqs=int(stats["n_seqs"]),
            )
        return res

    def save(self, model: model_api.Model, save_dir: str):
        if not self.enable_save:
            return
        module = model.module
        if not isinstance(module, ReaLModel):
            module = module.module
        module.save_to_hf(
            tokenizer=model.tokenizer,
            save_dir=save_dir,
        )


model_api.register_interface("dpo", DPOInterface)



from typing import Callable, Dict, List, Tuple
import ast
import copy
import enum
import hashlib
import itertools
import pickle
import time
import types
import warnings

import numpy as np
import torch

import realhf.base.logging as logging

sparse_tensor_fields = []


def dense_tensor_size(x: torch.Tensor):
    assert not x.is_sparse
    return x.numel() * x.element_size()


def sparse_tensor_size(x: torch.Tensor):
    assert x.is_sparse
    indices_size = x.indices().numel() * x.indices().element_size()
    values_size = x.values().numel() * x.values().element_size()
    return indices_size + values_size


class NamedArrayLoadingError(Exception):
    pass


class NamedArrayEncodingMethod(bytes, enum.Enum):
    """Encoding protocol of a NamedArray object.

    Principle: compress huge non-random data and pickle small data.

    TL;DR:
    + InferenceStream
        ++ If your observation include images, use PICKLE_COMPRESS.
        ++ Otherwise, use PICKLE or PICKLE_DICT.
    + SampleStream:
        ++ If your observation include images, use OBS_COMPRESS or COMPRESS_EXPECT_POLICY_STATE.
        ++ Otherwise,
            +++ if the amount of data is huge (say, multi-agent envs), use COMPRESS_EXPECT_POLICY_STATE.
            +++ otherwise, use PICKLE or PICKLE_DICT.
    """

    PICKLE_DICT = b"0001"  # Convert NamedArray to dict, then pickle.
    PICKLE = b"0002"  # Directly pickle.
    RAW_BYTES = b"0003"  # Send raw bytes of flattened numpy arrays.
    RAW_COMPRESS = b"0004"  # Send compressed bytes of flattened numpy arrays.
    COMPRESS_PICKLE = (
        b"0005"  # Convert numpy array to compressed bytes, then pickle.
    )
    PICKLE_COMPRESS = b"0006"  # Pickle, then compress pickled bytes.
    OBS_COMPRESS = b"0007"  # Convert flattened numpy array to bytes and only compress observation.
    COMPRESS_EXCEPT_POLICY_STATE = b"0008"  # Compress all bytes except for policy states, which are basically random numbers.
    TENSOR_COMPRESS = b"0009"  # Turn certain tensors into sparse tensors, then pickle. Turn other tensors into numpy arrays then compress.


logger = logging.getLogger("NamedArray")


def _namedarray_op(op):

    def fn(self, value):
        if not (
            isinstance(value, NamedArray)  # Check for matching structure.
            and getattr(value, "_fields", None) == self._fields
        ):
            if not isinstance(value, NamedArray):
                # Repeat value for each but respect any None.
                value = tuple(None if s is None else value for s in self)
            else:
                raise ValueError(
                    "namedarray - set an item with a different data structure"
                )
        try:
            xs = {}
            for j, ((k, s), v) in enumerate(zip(self.items(), value)):
                if s is not None and v is not None:
                    exec(f"xs[k] = (s {op} v)")
                else:
                    exec(f"xs[k] = None")
        except (ValueError, IndexError, TypeError) as e:
            print(s.shape, v.shape)
            raise Exception(
                f"{type(e).__name__} occured in {self.__class__.__name__}"
                " at field "
                f"'{self._fields[j]}': {e}"
            ) from e
        return NamedArray(**xs)

    return fn


def _namedarray_iop(iop):

    def fn(self, value):
        if not (
            isinstance(value, NamedArray)  # Check for matching structure.
            and getattr(value, "_fields", None) == self._fields
        ):
            if not isinstance(value, NamedArray):
                # Repeat value for each but respect any None.
                value = {
                    k: None if s is None else value for k, s in self.items()
                }
            else:
                raise ValueError(
                    "namedarray - set an item with a different data structure"
                )
        try:
            for j, (k, v) in enumerate(zip(self.keys(), value.values())):
                if self[k] is not None and v is not None:
                    exec(f"self[k] {iop} v")
        except (ValueError, IndexError, TypeError) as e:
            raise Exception(
                f"{type(e).__name__} occured in {self.__class__.__name__}"
                " at field "
                f"'{self._fields[j]}': {e}"
            ) from e
        return self

    return fn


def _numpy_dtype_to_str(dtype):
    if dtype == np.uint8:
        return "uint8"
    elif dtype == bool:
        return "bool"
    elif dtype == np.float32:
        return "float32"
    elif dtype == np.float16:
        return "float16"
    elif dtype == np.float64:
        return "float64"
    elif dtype == np.int32:
        return "int32"
    elif dtype == np.int64:
        return "int64"
    elif dtype == np.dtype("<U7"):
        return "<U7"
    else:
        raise NotImplementedError(
            f"Data type to string not implemented: {dtype}."
        )


def dumps(namedarray_obj, method="pickle_dict"):
    """Serialize a NamedArray object to bytes.

    Args:
        namedarray_obj (NamedArray): The NamedArray object to be serialized.
        method (str): The serialization method.
    """

    if "compress" in method:
        try:
            import blosc
        except ModuleNotFoundError:

            class blosc:

                def compress(x, *args, **kwargs):
                    return x

            warnings.warn(
                "Module `blosc` not found in the image. Abort NamedArray compression."
            )

    def compress_large(buf, typesize=4, cname="lz4"):
        chunk_size = 2**30
        num_chunks = len(buf) // chunk_size + 1
        chunks = [
            blosc.compress(
                buf[i * chunk_size : (i + 1) * chunk_size],
                typesize=typesize,
                cname=cname,
            )
            for i in range(num_chunks)
        ]
        # logger.info(f"Compress {len(buf)} bytes to {sum([len(chunk) for chunk in chunks])} bytes, num_chunks = {num_chunks}")
        return num_chunks, chunks

    def _namedarray_to_bytes_list(
        x, compress: bool, compress_condition: Callable[[str], bool]
    ):
        flattened_entries = flatten(x)
        flattened_bytes = []
        for k, v in flattened_entries:
            k_ = k.encode("ascii")
            dtype_ = v_ = shape_ = b""
            if v is not None:
                dtype_ = _numpy_dtype_to_str(v.dtype).encode("ascii")
                v_ = v.tobytes()
                shape_ = str(tuple(v.shape)).encode("ascii")
                if compress and compress_condition(k):
                    v_ = blosc.compress(v_, typesize=4, cname="lz4")
            flattened_bytes.append((k_, dtype_, shape_, v_))
        return list(itertools.chain.from_iterable(flattened_bytes))

    def _tensor_namedarray_to_bytes_list(x):
        t0 = time.monotonic()
        flattened_entries = flatten(x)
        flattened_bytes = []
        for k, v in flattened_entries:
            k_ = k.encode("ascii")
            dtype_ = v_ = shape_ = b""
            if k in sparse_tensor_fields:
                t1 = time.monotonic()
                assert v is not None, f"Sparse tensor field {k} cannot be None."
                logger.info(
                    f"Dense tensor {k} size {dense_tensor_size(v)} bytes"
                )
                vshape = v.shape
                v = v.to_sparse()
                logger.info(
                    f"Sparse tensor {k} size {sparse_tensor_size(v)} bytes"
                )
                v_ = pickle.dumps(v)
                logger.info(
                    f"dump sparse tensor {k} time {time.monotonic() - t1:4f}, v shape {vshape}"
                )
            elif v is not None:
                v = v.cpu().numpy()
                dtype_ = _numpy_dtype_to_str(v.dtype).encode("ascii")
                v_ = v.tobytes()
                shape_ = str(tuple(v.shape)).encode("ascii")
                v_ = blosc.compress(v_, typesize=4, cname="lz4")
            flattened_bytes.append((k_, dtype_, shape_, v_))
        # logger.info(f"dump namedarray time {time.monotonic() - t0:4f}")
        return list(itertools.chain.from_iterable(flattened_bytes))

    if method == "pickle_dict":
        bytes_list = [
            NamedArrayEncodingMethod.PICKLE_DICT.value,
            pickle.dumps(
                (namedarray_obj.__class__.__name__, namedarray_obj.to_dict())
            ),
        ]
    elif method == "pickle":
        bytes_list = [
            NamedArrayEncodingMethod.PICKLE.value,
            pickle.dumps(namedarray_obj),
        ]
    elif method == "raw_bytes":
        bytes_list = [
            NamedArrayEncodingMethod.RAW_BYTES.value
        ] + _namedarray_to_bytes_list(namedarray_obj, False, lambda x: False)
    elif method == "raw_compress":
        bytes_list = [
            NamedArrayEncodingMethod.RAW_COMPRESS.value
        ] + _namedarray_to_bytes_list(namedarray_obj, True, lambda x: True)
    elif method == "compress_pickle":
        bytes_list = [
            NamedArrayEncodingMethod.COMPRESS_PICKLE.value,
            pickle.dumps(
                _namedarray_to_bytes_list(namedarray_obj, True, lambda x: True)
            ),
        ]
    elif method == "pickle_compress":
        # bytes_list = [
        #     NamedArrayEncodingMethod.PICKLE_COMPRESS.value,
        #     blosc.compress(pickle.dumps(namedarray_obj), typesize=4, cname='lz4')
        # ]
        bytes_list = [NamedArrayEncodingMethod.PICKLE_COMPRESS.value]
        buf = pickle.dumps(namedarray_obj)
        num_chunks, chunks = compress_large(buf)
        bytes_list += chunks
    elif method == "obs_compress":
        bytes_list = [
            NamedArrayEncodingMethod.OBS_COMPRESS.value
        ] + _namedarray_to_bytes_list(
            namedarray_obj, True, lambda x: ("obs" in x)
        )
    elif method == "compress_except_policy_state":
        bytes_list = [
            NamedArrayEncodingMethod.COMPRESS_EXCEPT_POLICY_STATE.value
        ] + _namedarray_to_bytes_list(
            namedarray_obj, True, lambda x: ("policy_state" not in x)
        )

    elif method == "tensor_compress":
        bytes_list = [
            NamedArrayEncodingMethod.TENSOR_COMPRESS.value
        ] + _tensor_namedarray_to_bytes_list(namedarray_obj)
    else:
        raise NotImplementedError(
            f"Unknown method {method}. Available are {[m.name.lower() for m in NamedArrayEncodingMethod]}."
        )

    return bytes_list + [pickle.dumps(dict(**namedarray_obj.metadata))]


def loads(b):
    # safe import
    if b[0] in [b"0004", b"0004", b"0005", b"0006", b"0007", b"0008", b"0009"]:
        try:
            import blosc
        except ModuleNotFoundError:

            class blosc:

                def decompress(x, *args, **kwargs):
                    return x

    def _parse_namedarray_from_bytes_list(
        xs, compressed: bool, compress_condition: Callable[[str], int]
    ):
        flattened = []
        for i in range(len(xs) // 4):
            k = xs[4 * i].decode("ascii")
            if xs[4 * i + 1] != b"":
                buf = xs[4 * i + 3]
                if compressed and compress_condition(k):
                    buf = blosc.decompress(buf)
                v = np.frombuffer(
                    buf, dtype=np.dtype(xs[4 * i + 1].decode("ascii"))
                ).reshape(*ast.literal_eval(xs[4 * i + 2].decode("ascii")))
            else:
                v = None
            flattened.append((k, v))
        return from_flattened(flattened)

    def _parse_tensor_namedarray_from_bytes_list(xs):
        t0 = time.monotonic()
        flattened = []
        for i in range(len(xs) // 4):
            k = xs[4 * i].decode("ascii")
            if xs[4 * i + 3] == b"":
                # None
                v = None
            elif xs[4 * i + 1] == b"":
                t1 = time.monotonic()
                # sparse tensor
                v = pickle.loads(xs[4 * i + 3])
                assert (
                    torch.is_tensor(v) and v.is_sparse
                ), f"Field {k} is not a sparse tensor, but is serialized as one."
                logger.info(
                    f"Sparse tensor {k} size {sparse_tensor_size(v)} bytes"
                )
                v = v.to_dense()
                logger.info(
                    f"Dense tensor {k} size {dense_tensor_size(v)} bytes"
                )
                logger.info(
                    f"load sparse tensor {k} time {time.monotonic() - t1:4f}, v shape {v.shape}"
                )
            else:
                # dense tensor
                buf = xs[4 * i + 3]
                buf = blosc.decompress(buf)
                v = np.frombuffer(
                    buf, dtype=np.dtype(xs[4 * i + 1].decode("ascii"))
                ).reshape(*ast.literal_eval(xs[4 * i + 2].decode("ascii")))
                v = torch.from_numpy(v)
            flattened.append((k, v))
        # logger.info(f"load namedarray time {time.monotonic() - t0:4f}")
        return from_flattened(flattened)

    if b[0] == NamedArrayEncodingMethod.PICKLE_DICT.value:
        class_name, values = pickle.loads(b[1])
        namedarray_obj = from_dict(values=values)
    elif b[0] == NamedArrayEncodingMethod.PICKLE.value:
        namedarray_obj = pickle.loads(b[1])
    elif b[0] == NamedArrayEncodingMethod.RAW_BYTES.value:
        namedarray_obj = _parse_namedarray_from_bytes_list(
            b[1:-1], False, lambda x: False
        )
    elif b[0] == NamedArrayEncodingMethod.RAW_COMPRESS.value:
        namedarray_obj = _parse_namedarray_from_bytes_list(
            b[1:-1], True, lambda x: True
        )
    elif b[0] == NamedArrayEncodingMethod.COMPRESS_PICKLE.value:
        namedarray_obj = _parse_namedarray_from_bytes_list(
            pickle.loads(b[1]), True, lambda x: True
        )
    elif b[0] == NamedArrayEncodingMethod.PICKLE_COMPRESS.value:
        # namedarray_obj = pickle.loads(blosc.decompress(b[1]))
        chunks = b[1:-1]
        buf = b"".join([blosc.decompress(chunk) for chunk in chunks])
        namedarray_obj = pickle.loads(buf)
    elif b[0] == NamedArrayEncodingMethod.OBS_COMPRESS.value:
        namedarray_obj = _parse_namedarray_from_bytes_list(
            b[1:-1], True, lambda x: ("obs" in x)
        )
    elif b[0] == NamedArrayEncodingMethod.COMPRESS_EXCEPT_POLICY_STATE.value:
        namedarray_obj = _parse_namedarray_from_bytes_list(
            b[1:-1], True, lambda x: ("policy_state" not in x)
        )
    elif b[0] == NamedArrayEncodingMethod.TENSOR_COMPRESS.value:
        namedarray_obj = _parse_tensor_namedarray_from_bytes_list(b[1:-1])
    else:
        raise NotImplementedError(
            f"Unknown NamedArrayEncodingMethod value {b[:4]}. "
            f"Existing are {[m for m in NamedArrayEncodingMethod]}."
        )

    namedarray_obj.clear_metadata()
    metadata = pickle.loads(b[-1])
    namedarray_obj.register_metadata(**metadata)

    return namedarray_obj


class NamedArray:
    """A class acting like a dict that holds torch tensors as values.

    NamedArray supports dict-like unpacking and string indexing, and exposes integer slicing reads
    and writes applied to all contained objects, which must share
    indexing (__getitem__) behavior (e.g. numpy arrays or torch tensors).

    NamedArray also supports metadata, which is useful for recording sequence lengths.

    Example:

    >>> class Point(NamedArray):
    ...     def __init__(self,
    ...         x: np.ndarray,
    ...         y: np.ndarray,
    ...         ):
    ...         super().__init__(x=x, y=y)
    >>> p=Point(np.array([1,2]), np.array([3,4]))
    >>> p
    Point(x=array([1, 2]), y=array([3, 4]))
    >>> p[:-1]
    Point(x=array([1]), y=array([3]))
    >>> p[0]
    Point(x=1, y=3)
    >>> p.x
    array([1, 2])
    >>> p['y']
    array([3, 4])
    >>> p[0] = 0
    >>> p
    Point(x=array([0, 2]), y=array([0, 4]))
    >>> p[0] = Point(5, 5)
    >>> p
    Point(x=array([5, 2]), y=array([5, 4]))
    >>> 'x' in p
    True
    >>> list(p.keys())
    ['x', 'y']
    >>> list(p.values())
    [array([5, 2]), array([5, 4])]
    >>> for k, v in p.items():
    ...     print(k, v)
    ...
    x [5 2]
    y [5 4]
    >>> def foo(x, y):
    ...     print(x, y)
    ...
    >>> foo(**p)
    [5 2] [5 4]
    """

    _reserved_slots = ["_NamedArray__metadata", "_fields"]

    def __init__(self, **kwargs):
        self._fields = list(sorted(kwargs.keys()))
        self.__metadata = types.MappingProxyType({})
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def metadata(self):
        """Return the metadata of the NamedArray object.

        :return: A dict-like metadata of the NamedArray object.
        :rtype: MappingProxyType
        """
        return self.__metadata

    def register_metadata(self, **kwargs):
        """Register metadata to the NamedArray object.

        :param kwargs: Metadata to be registered.
        :type kwargs: Dict
        """
        for k in self._fields:
            if k in kwargs.keys():
                raise KeyError(
                    "Keys of metadata should be different from data fields!"
                )

        self.__metadata = types.MappingProxyType({**self.__metadata, **kwargs})

    def pop_metadata(self, key):
        """Clear a single metadata entry named "key".

        :param key: The key to be removed.
        :type key: str
        :return: The value of the removed key.
        :rtype: Any
        """
        metadatadict = dict(self.__metadata)
        value = metadatadict.pop(key)
        self.__metadata = types.MappingProxyType(metadatadict)
        return value

    def clear_metadata(self):
        """Clear all metadata recorded in this object."""
        self.__metadata = types.MappingProxyType({})

    def __iter__(self):
        for k in self._fields:
            yield getattr(self, k)

    def __setattr__(self, loc, value):
        """Set attributes in a `NamedArray` object.

        Unknown fields cannot be created.

        :param loc: The attribute name to be set.
        :type loc: str or slice
        :param value: The value to be set.
        :type value: Any
        """
        if not (loc in NamedArray._reserved_slots or loc in self._fields):
            self._fields.append(loc)
        super().__setattr__(loc, value)

    def __getitem__(self, loc):
        """If the index is string, return getattr(self, index).
        If the index is integer/slice, return a new dataclass instance containing
        the selected index or slice from each field.

        :param loc: Key or indices to get.
        :type loc: str or slice
        :return: An element of NamedArray or a new NamedArray
            object composed of the slices.
        :rtype: NamedArray or Any
        """
        if isinstance(loc, str):
            # str indexing like in dict
            return getattr(self, loc)
        else:
            sliced_namedarray = dict()
            try:
                for s in self._fields:
                    if self[s] is None:
                        sliced_namedarray[s] = None
                    else:
                        sliced_namedarray[s] = self[s][loc]
            except IndexError as e:
                raise Exception(
                    f"IndexError occured when slicing `NamedArray`."
                    f"Field {s} with shape {self[s].shape} and slice {loc}."
                ) from e
            return self.__class__(
                **{
                    s: None if self[s] is None else self[s][loc]
                    for s in self._fields
                }
            )

    def __setitem__(self, loc, value):
        """If input value is the same dataclass type, iterate through its
        fields and assign values into selected index or slice of corresponding
        field.  Else, assign whole of value to selected index or slice of
        all fields. Ignore fields that are both None.

        :param loc: Key or indices to set.
        :type loc: str or slice
        :param value: An NamedArray instance with the same structure
            or elements of the NamedArray object.
        :type value: Any
        """
        if isinstance(loc, str):
            if loc not in self._fields:
                self._fields.append(loc)
            setattr(self, loc, value)
        else:
            if not (
                isinstance(value, NamedArray)  # Check for matching structure.
                and getattr(value, "_fields", None) == self._fields
            ):
                if not isinstance(value, NamedArray):
                    # Repeat value for each but respect any None.
                    value = tuple(None if s is None else value for s in self)
                else:
                    raise ValueError(
                        "namedarray - set an item with a different data structure"
                    )
            try:
                for j, (s, v) in enumerate(zip(self, value)):
                    if s is not None and v is not None:
                        s[loc] = v
            except (ValueError, IndexError, TypeError) as e:
                raise Exception(
                    f"Error occured occured in {self.__class__.__name__} when assigning value"
                    " at field "
                    f"'{self._fields[j]}': {e}"
                ) from e

    def __contains__(self, key):
        """Checks presence of a field name (unlike tuple; like dict).

        :param key: The queried field name.
        :type key: str
        :return: Query result.
        :rtype: bool
        """
        return key in self._fields

    def __getstate__(self):
        return {
            "__metadata": dict(**self.metadata),
            **{k: v for k, v in self.items()},
        }

    def __setstate__(self, state):
        self.__init__(**{k: v for k, v in state.items() if k != "__metadata"})
        if state["__metadata"] is not None:
            self.clear_metadata()
            self.register_metadata(**state["__metadata"])

    def values(self):
        for v in self:
            yield v

    def keys(self):
        for k in self._fields:
            yield k

    def __len__(self):
        return len(self._fields)

    def length(self, dim=0):
        for k, v in self.items():
            if v is None:
                continue
            elif isinstance(v, (np.ndarray, torch.Tensor)):
                if dim < v.ndim:
                    return v.shape[dim]
                else:
                    continue
            else:
                continue
        else:
            raise IndexError(f"No entry has shape on dim={dim}.")

    def unique_of(self, field, exclude_values=(None,)):
        unique_values = np.unique(self[field])
        unique_values = unique_values[
            np.in1d(unique_values, exclude_values, invert=True)
        ]
        if len(unique_values) != 1:
            return None
        else:
            return unique_values[0]

    def items(self):
        """Iterate over ordered (field_name, value) pairs like a dict."""
        for k, v in zip(self._fields, self):
            yield k, v

    def to_dict(self):
        result = {}
        for k, v in self.items():
            if isinstance(v, NamedArray):
                result[k] = v.to_dict()
            elif v is None:
                result[k] = None
            else:
                result[k] = v
        return result

    @property
    def shape(self):
        return recursive_apply(self, lambda x: x.shape).to_dict()

    def size(self):
        return self.shape

    def __str__(self):
        return f"{self.__class__.__name__}({', '.join(k+'='+repr(v) for k, v in self.items())})"

    def __hash__(self):
        d = self.to_dict()
        to_hash = []
        for k, v in d.items():
            if torch.is_tensor(v):
                vv = int(hashlib.md5(v.numpy().tobytes()).hexdigest(), 16)
            else:
                vv = hash(v)
            to_hash.append(k)
            to_hash.append(vv)
        serialized = pickle.dumps(to_hash)
        return int(hashlib.md5(serialized).hexdigest(), 16)

    def __deepcopy__(self, memo={}):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        metadata = {}
        for k, v in self.__dict__.items():
            if isinstance(v, types.MappingProxyType):
                metadata = copy.deepcopy(dict(v), memo)
                continue
            else:
                setattr(result, k, copy.deepcopy(v, memo))
        result.clear_metadata()
        result.register_metadata(**metadata)
        return result

    __add__ = _namedarray_op("+")
    __sub__ = _namedarray_op("-")
    __mul__ = _namedarray_op("*")
    __truediv__ = _namedarray_op("/")
    __iadd__ = _namedarray_iop("+=")
    __isub__ = _namedarray_iop("-=")
    __imul__ = _namedarray_iop("*=")
    __itruediv__ = _namedarray_iop("/=")
    __repr__ = __str__


def from_dict(values: Dict):
    """Create namedarray object from Nested Dict of arrays.

    Example:
    >>> a = from_dict({"x": np.array([1, 2]), "y": np.array([3,4])})
    >>> a.x
    array([1, 2])
    >>> a.y
    array([3, 4])
    >>> a[1:]
    NamedArray(x=[2],y=[4])
    >>> obs = {"state":{"speed": np.array([1, 2, 3]), "position": np.array([4, 5])}, "vision": np.array([[7],[8],[9]])}
    >>> obs_na = from_dict(obs)
    >>> obs_na
    NamedArray(state=NamedArray(position=[4 5],speed=[1 2 3]),vision=[[7]
     [8]
     [9]])
    >>> obs_na.state
    NamedArray(position=[4 5],speed=[1 2 3])

    :param values: Nested key-value object of data. value should of type None, Numpy Array, or Torch Tensor.
    :type values: Dict
    :return: NamedArray with the same data structure as input. If values is empty, return None.
    :rtype: NamedArray
    """
    if values is None or len(values) == 0:
        return None
    for k, v in values.items():
        if isinstance(v, dict):
            values[k] = from_dict(v)
    return NamedArray(**values)


def array_like(x, value=0):
    if isinstance(x, NamedArray):
        return NamedArray(**{k: array_like(v, value) for k, v in x.items()})
    else:
        if isinstance(x, np.ndarray):
            data = np.zeros_like(x)
        else:
            assert isinstance(x, torch.Tensor), (
                "Currently, namedarray only supports"
                f" torch.Tensor and numpy.array (input is {type(x)})"
            )
            data = torch.zeros_like(x)
        if value != 0:
            data[:] = value
        return data


def __array_filter_none(xs):
    is_not_nones = [x is not None for x in xs]
    if all(is_not_nones) or all(x is None for x in xs):
        return
    else:
        example_x = xs[is_not_nones.index(True)]
        for i, x in enumerate(xs):
            xs[i] = array_like(example_x) if x is None else x


def recursive_aggregate(xs, aggregate_fn):
    """Recursively aggregate a list of namedarray instances.
    Typically recursively stacking or concatenating.

    :param xs: A list of NamedArrays or appropriate aggregation targets (e.g. numpy.ndarray).
    :type xs: List[NamedArray or Any]
    :param aggregate_fn: The aggregation function to be applied.
    :type aggregate_fn: Callable
    :return: The aggregated result with the same data type of elements in xs.
    :rtype: NamedArray or Any
    """
    __array_filter_none(xs)
    if isinstance(xs[0], NamedArray):
        entries = dict()
        for k in xs[0].keys():
            try:
                entries[k] = recursive_aggregate(
                    [x[k] for x in xs], aggregate_fn
                )
            except Exception as e:
                err_msg = (
                    f"`recursive_aggregate` fails at an entry named `{k}`."
                )
                if not all([type(x[k]) == type(xs[0][k]) for x in xs]):
                    err_msg += f" Types of elements are not the same: {[type(x[k]) for x in xs]}."
                else:
                    if isinstance(xs[0][k], NamedArray):
                        err_msg += " Elements are all `NamedArray`s. Backtrace to the above level."
                if not any([isinstance(x[k], NamedArray) for x in xs]):
                    specs = []
                    for x in xs:
                        specs.append(f"({x[k].dtype}, {tuple(x[k].shape)})")
                    err_msg += f" Specs of elements to be aggregated are: [{', '.join(specs)}]."
                raise RuntimeError(err_msg) from e
        return NamedArray(**entries)
    elif xs[0] is None:
        return None
    else:
        return aggregate_fn(xs)


def recursive_apply(x, fn):
    """Recursively apply a function to an NamedArray x.

    :param x: The instance of a namedarray subclass or an appropriate target to apply fn.
    :type x: NamedArray or Any
    :param fn: The function to be applied.
    :type fn: Callable
    :return: The result of applying fn to x, preserving metadata.
    :rtype: NamedArray or Any
    """
    if isinstance(x, NamedArray):
        entries = dict()
        for k, v in x.items():
            try:
                entries[k] = recursive_apply(v, fn)
            except Exception as e:
                err_msg = f"`recursive_apply` fails at an entry named `{k}`"
                if isinstance(v, NamedArray):
                    err_msg += ", which is a `NamedArray`. Backtrace to the above level."
                else:
                    err_msg += f" ({v.dtype}, {tuple(v.shape)})."
                raise RuntimeError(err_msg) from e
        res = NamedArray(**entries)
        res.register_metadata(**x.metadata)
        return res
    elif x is None:
        return None
    else:
        return fn(x)


def flatten(x: NamedArray) -> List[Tuple]:
    """Flatten a NamedArray object to a list containing structured names and values."""
    flattened_entries = []
    for k, v in x.items():
        if isinstance(v, NamedArray):
            flattened_entries += [(f"{k}." + k_, v_) for k_, v_ in flatten(v)]
        else:
            flattened_entries.append((k, v))
    return flattened_entries


def from_flattened(flattened_entries):
    """Construct a NamedArray from flattened names and values."""
    keys, values = zip(*flattened_entries)
    entries = dict()
    idx = 0
    while idx < len(keys):
        k, v = keys[idx], values[idx]
        if "." not in k:
            entries[k] = v
            idx += 1
        else:
            prefix = k.split(".")[0]
            span_end = idx + 1
            while span_end < len(keys) and keys[span_end].startswith(
                f"{prefix}."
            ):
                span_end += 1
            subentries = [
                (keys[j][len(prefix) + 1 :], values[j])
                for j in range(idx, span_end)
            ]
            entries[prefix] = from_flattened(subentries)
            idx = span_end
    return from_dict(entries)


def merge(xs: List[NamedArray]):
    """merge named arrays with different field into one, identical fields
    will be overwritten by the last one in the list
    """
    mixed_dict = {}
    for x in xs:
        mixed_dict.update(x.to_dict())
    return from_dict(mixed_dict)


def split(x: NamedArray, n):
    """split a named array into n parts along dim 0"""
    sz = x.length(0) // n
    return [x[sz * i : sz * (i + 1)] for i in range(n - 1)] + [
        x[sz * (n - 1) :]
    ]



from typing import Dict, List, Optional, Union
import getpass
import json
import os
import re
import socket
import tempfile

CLUSTER_SPEC_PATH = os.environ.get("CLUSTER_SPEC_PATH", "")


def get_user_tmp():
    tmp = tempfile.gettempdir()
    user = getpass.getuser()
    user_tmp = os.path.join(tmp, user)
    os.makedirs(user_tmp, exist_ok=True)
    return user_tmp


def get_random_tmp():
    return tempfile.mkdtemp()


class ClusterSpec:
    def __init__(self):
        self.__loaded = False

    def load_spec_from_file(self, file_path: str):
        try:
            with open(file_path, "r") as f:
                spec: Dict = json.load(f)
        except FileNotFoundError:
            if file_path == "":
                spec = dict(
                    cluster_type="local",
                    cluster_name="local",
                    fileroot=get_user_tmp(),
                )
            else:
                raise FileNotFoundError(
                    f"Cluster spec file not found: {file_path}"
                )

        self.__cluster_type = spec["cluster_type"]
        self.__cluster_name = spec["cluster_name"]
        self.__fileroot = spec["fileroot"]
        self.__node_type_from_node_name_re = spec.get(
            "node_type_from_node_name", None
        )
        self.__gpu_type_from_node_name_re = spec.get(
            "gpu_type_from_node_name", None
        )
        self.__default_mount = spec.get("default_mount", None)
        self.__gpu_image = spec.get("gpu_image", None)
        self.__cpu_image = spec.get("cpu_image", None)
        self.__node_name_prefix = spec.get("node_name_prefix", "NODE")

        self.__loaded = True

    @property
    def name(self):
        assert self.__loaded
        return self.__cluster_name

    def node_type_from_node_name(self, node_name: str) -> str:
        """mapping nodename to slurm node type, including "g1", "g2", "g8", "a100" """
        if self.__cluster_type != "slurm":
            raise NotImplementedError(
                "Only slurm cluster uses node_type_from_node_name."
            )
        assert self.__loaded
        for regex, node_type in self.__node_type_from_node_name_re.items():
            if re.match(regex, node_name):
                return node_type
        raise NotImplementedError()

    def gpu_type_from_node_name(self, node_name: str) -> str:
        """mapping nodename to slurm GPU type, including "geforce" and "tesla" """
        if self.__cluster_type != "slurm":
            raise NotImplementedError(
                "Only slurm cluster uses gpu_type_from_node_name."
            )
        assert self.__loaded
        for regex, gpu_type in self.__gpu_type_from_node_name_re.items():
            if re.match(regex, node_name):
                return gpu_type
        raise NotImplementedError()

    @property
    def fileroot(self) -> str:
        """Return the root directory of the file system in the cluster.
        When running experiments, files such as logs, checkpoints, caches
        will be saved under this directory.
        """
        assert self.__loaded
        return self.__fileroot

    @property
    def default_mount(self) -> str:
        """Directories that should be mounted to container that runs workers."""
        assert self.__loaded
        return self.__default_mount

    @property
    def gpu_image(self) -> str:
        """Return the default image for containers of GPU workers."""
        assert self.__loaded
        return self.__gpu_image

    @property
    def cpu_image(self) -> str:
        """Return the default image for containers of CPU workers."""
        assert self.__loaded
        return self.__cpu_image

    @property
    def node_name_prefix(self) -> str:
        """Return the prefix of node names in slurm format"""
        assert self.__loaded
        return self.__node_name_prefix


def node_name_is_node_type(
    node_name: str, node_type: Optional[Union[List[str], str]] = None
) -> bool:
    assert spec is not None
    if node_type is None:
        return True
    if not isinstance(node_type, list):
        node_type = [node_type]
    nt_condition = []
    for nt in node_type:
        if nt not in ["g1", "g2", "g8", "a100"]:
            raise ValueError(f"Unknown node type {nt}.")
        else:
            cond = spec.node_type_from_node_name(node_name) == nt
        nt_condition.append(cond)
    return any(nt_condition)


spec = ClusterSpec()
spec.load_spec_from_file(CLUSTER_SPEC_PATH)



from typing import *
import dataclasses
import functools
import math

from deepspeed.runtime import zero
from deepspeed.runtime.bf16_optimizer import BF16_Optimizer
from deepspeed.runtime.config import DeepSpeedConfig
from deepspeed.runtime.engine import (
    DeepSpeedEngine,
    DeepSpeedOptimizerCallable,
    DeepSpeedSchedulerCallable,
)
import deepspeed
import torch
import torch.distributed as dist
import transformers

from realhf.impl.model.backend.pipe_runner import PipelineRunner
from realhf.impl.model.nn.real_llm_api import ReaLModel
from realhf.impl.model.nn.real_llm_generate import GenerationConfig
import realhf.api.core.model_api as model_api
import realhf.base.constants as constants
import realhf.base.logging as logging

logger = logging.getLogger("DeepSpeed Backend")


class ReaLDeepSpeedEngine:

    def __init__(
        self,
        module: ReaLModel,
        ds_engine: DeepSpeedEngine,
    ):
        self.module = module

        self.device = module.device
        self.dtype = module.dtype

        self.ds_engine = ds_engine

        if constants.pipe_parallel_world_size() > 1:
            self.pipe_runner = PipelineRunner(module)

            assert (
                self.ds_engine.zero_optimization_stage() < 2
            ), "ZeRO-2 and ZeRO-3 are incompatible with pipeline parallelism"
            if self.ds_engine.bfloat16_enabled():
                assert isinstance(self.ds_engine.optimizer, BF16_Optimizer)

            model_parameters = filter(
                lambda p: p.requires_grad, self.module.parameters()
            )
            num_params = sum([p.numel() for p in model_parameters])
            unique_params = num_params

            params_tensor = torch.LongTensor(
                data=[num_params, unique_params]
            ).to(self.device)
            dist.all_reduce(
                params_tensor, group=constants.grid().get_model_parallel_group()
            )
            params_tensor = params_tensor.tolist()
            total_params = params_tensor[0]
            unique_params = params_tensor[1]

            if constants.parallelism_rank() == 0:
                logger.info(
                    f"CONFIG: default_train_mbs={self.pipe_runner.default_train_mbs} "
                    f"default_inf_mbs={self.pipe_runner.default_inf_mbs} "
                    f"num_layers(this stage)={self.module.num_layers} "
                    f"pp_size={constants.pipe_parallel_world_size()} "
                    f"dp_size={constants.data_parallel_world_size()} "
                    f"mp_size={constants.model_parallel_world_size()} "
                    f"bf16={self.ds_engine.bfloat16_enabled()} "
                )
            if constants.data_parallel_rank() == 0:
                logger.info(
                    f"rank={constants.parallelism_rank()} "
                    f"stage={constants.pipe_parallel_rank()} "
                    f"layers={self.module.num_layers} "
                    f"[{self.module.layer_idx_start}, {self.module.layer_idx_end}) "
                    f"stage_params={num_params} ({num_params/1e6:0.3f}M) "
                    f"total_params={total_params} ({total_params/1e6:0.3f}M) "
                    f"unique_params={unique_params} ({unique_params/1e6:0.3f}M)"
                )

    def train(self, mode: bool = True):
        self.ds_engine.train(mode)
        self.module.train(mode)
        return self

    def eval(self):
        self.ds_engine.eval()
        self.module.eval()
        return self

    def train_batch(
        self,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        loss_fn: Callable,
        version_steps: int,
        input_lens_for_partition: Optional[torch.Tensor] = None,
        num_micro_batches: Optional[int] = None,
        **loss_fn_kwargs,
    ):
        if constants.pipe_parallel_world_size() > 1:
            return self.pipe_runner.train_batch(
                engine=self.ds_engine,
                seqlens_cpu=seqlens_cpu,
                packed_input_ids=packed_input_ids,
                cu_seqlens=cu_seqlens,
                loss_fn=loss_fn,
                version_steps=version_steps,
                input_lens_for_partition=input_lens_for_partition,
                num_micro_batches=num_micro_batches,
                **loss_fn_kwargs,
            )
        else:
            # FIXME: num_micro_batches is ignored here
            max_seqlen = int(max(cu_seqlens[1:] - cu_seqlens[:-1]))
            model_output = self.ds_engine(
                packed_input_ids=packed_input_ids,
                cu_seqlens=cu_seqlens,
                max_seqlen=max_seqlen,
            ).logits
            loss, stat = loss_fn(
                model_output, packed_input_ids, cu_seqlens, **loss_fn_kwargs
            )
            self.ds_engine.backward(loss)
            lr_kwargs = (
                {"epoch": version_steps} if version_steps is not None else None
            )
            self.ds_engine.step(lr_kwargs=lr_kwargs)
            return stat

    @torch.no_grad()
    def eval_batch(
        self,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        loss_fn: Callable,
        input_lens_for_partition: Optional[torch.Tensor] = None,
        num_micro_batches: Optional[int] = None,
        **loss_fn_kwargs,
    ):
        if constants.pipe_parallel_world_size() > 1:
            return self.pipe_runner.eval_batch(
                seqlens_cpu=seqlens_cpu,
                packed_input_ids=packed_input_ids,
                cu_seqlens=cu_seqlens,
                loss_fn=loss_fn,
                input_lens_for_partition=input_lens_for_partition,
                num_micro_batches=num_micro_batches,
                **loss_fn_kwargs,
            )
        else:
            max_seqlen = int(max(cu_seqlens[1:] - cu_seqlens[:-1]))
            model_output = self.ds_engine(
                packed_input_ids=packed_input_ids,
                cu_seqlens=cu_seqlens,
                max_seqlen=max_seqlen,
            ).logits
            _, stat = loss_fn(
                model_output, packed_input_ids, cu_seqlens, **loss_fn_kwargs
            )
            return stat

    def forward(
        self,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        num_micro_batches: Optional[int] = None,
    ):
        if constants.pipe_parallel_world_size() > 1:
            return self.pipe_runner.forward(
                packed_input_ids=packed_input_ids,
                cu_seqlens=cu_seqlens,
                seqlens_cpu=seqlens_cpu,
                num_micro_batches=num_micro_batches,
            )
        else:
            max_seqlen = int(max(cu_seqlens[1:] - cu_seqlens[:-1]))
            return self.ds_engine(
                packed_input_ids=packed_input_ids,
                cu_seqlens=cu_seqlens,
                max_seqlen=max_seqlen,
            ).logits

    @torch.no_grad()
    def generate(
        self,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        tokenizer: transformers.PreTrainedTokenizerFast,
        gconfig: GenerationConfig = dataclasses.field(
            default_factory=GenerationConfig
        ),
        num_micro_batches: Optional[int] = None,
    ):
        if constants.pipe_parallel_world_size() > 1:
            return self.pipe_runner.generate(
                seqlens_cpu=seqlens_cpu,
                num_micro_batches=num_micro_batches,
                tokenizer=tokenizer,
                packed_input_ids=packed_input_ids,
                cu_seqlens=cu_seqlens,
                gconfig=gconfig,
            )
        else:
            max_seqlen = int(max(cu_seqlens[1:] - cu_seqlens[:-1]))
            res = self.module.generate(
                tokenizer=tokenizer,
                packed_input_ids=packed_input_ids,
                cu_seqlens=cu_seqlens,
                max_seqlen=max_seqlen,
                gconfig=gconfig,
            )
            return res.sequences, res.scores, res.logits_mask


def get_train_ds_config(
    offload_param: bool = False,
    offload_optimizer_state: bool = False,
    enable_fp16: bool = True,
    enable_bf16: bool = False,
    stage: int = 2,
    **kwargs,
):
    if enable_bf16 and enable_fp16:
        raise ValueError("Cannot enable both fp16 and bf16 at the same time.")
    zero_opt_dict = {
        "stage": stage,
        "overlap_comm": True,
        "round_robin_gradients": True,
        "offload_param": {
            "device": "cpu" if offload_param else "none",
            "pin_memory": True,
        },
        "offload_optimizer": {
            "device": "cpu" if offload_optimizer_state else "none",
            "pin_memory": True,
        },
        "stage3_param_persistence_threshold": 1e4,
        "stage3_max_live_parameters": 3e7,
        "stage3_prefetch_bucket_size": 3e7,
        "memory_efficient_linear": False,
    }
    return {
        "steps_per_print": 100,
        "zero_optimization": zero_opt_dict,
        "fp16": {
            "enabled": enable_fp16,
            "loss_scale_window": 40,
            "initial_scale_power": 12,
        },
        "bf16": {
            "enabled": enable_bf16,
        },
        "data_types": {
            "grad_accum_dtype": "fp32" if enable_bf16 else "fp16",
        },
        "gradient_clipping": 1.0,
        "prescale_gradients": False,
        "gradient_predevide_factor": 1.0,
        "wall_clock_breakdown": False,
        **kwargs,
    }


def get_eval_ds_config(
    offload=False,
    stage=0,
    enable_fp16: bool = True,
    enable_bf16: bool = False,
    **kwargs,
):
    if enable_bf16 and enable_fp16:
        raise ValueError("Cannot enable both fp16 and bf16 at the same time.")
    device = "cpu" if offload else "none"
    zero_opt_dict = {
        "stage": stage,
        "stage3_param_persistence_threshold": 1e4,
        "offload_param": {
            "device": device,
            "pin_memory": True,
        },
        "memory_efficient_linear": False,
    }
    return {
        "steps_per_print": 10,
        "zero_optimization": zero_opt_dict,
        "fp16": {
            "enabled": enable_fp16,
        },
        "bf16": {
            "enabled": enable_bf16,
        },
        "gradient_clipping": 1.0,
        "prescale_gradients": False,
        "wall_clock_breakdown": False,
        **kwargs,
    }


def get_optimizer_grouped_parameters(
    model: torch.nn.Module,
    weight_decay: float,
    no_decay_name_list: List[str] = ["bias", "ln.weight", "ln_f.weight"],
):
    optimizer_grouped_parameters = [
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if (
                    not any(nd in n for nd in no_decay_name_list)
                    and p.requires_grad
                )
            ],
            "weight_decay": weight_decay,
        },
        {
            "params": [
                p
                for n, p in model.named_parameters()
                if (
                    any(nd in n for nd in no_decay_name_list)
                    and p.requires_grad
                )
            ],
            "weight_decay": 0.0,
        },
    ]
    return optimizer_grouped_parameters


def deepspeed_initialize(
    model: torch.nn.Module,
    config: Dict,
    optimizer: Optional[
        Union[torch.optim.Optimizer, DeepSpeedOptimizerCallable]
    ] = None,
    model_parameters: Optional[torch.nn.Module] = None,
    lr_scheduler: Optional[
        Union[torch.optim.lr_scheduler._LRScheduler, DeepSpeedSchedulerCallable]
    ] = None,
    mpu=None,
) -> Tuple[DeepSpeedEngine, torch.optim.Optimizer, Any, Any]:
    """A simple wrapper around deepspeed.initialize."""
    if mpu is None:
        mpu = constants.grid()
    config_class = DeepSpeedConfig(config, mpu)
    logger.info(
        f"DeepSpeedEngine Config: train_batch_size={config_class.train_batch_size}, "
        f"train_micro_batch_size_per_gpu={config_class.train_micro_batch_size_per_gpu}, "
        f"gradient_accumulation_steps={config_class.gradient_accumulation_steps}"
    )

    # Disable zero.Init context if it's currently enabled
    zero.partition_parameters.shutdown_init_context()

    from deepspeed import comm as dist

    deepspeed.dist = dist

    engine = DeepSpeedEngine(
        args=None,
        model=model,
        optimizer=optimizer,
        model_parameters=model_parameters,
        lr_scheduler=lr_scheduler,
        mpu=mpu,
        dist_init_required=False,
        config=config,
        config_class=config_class,
        # dont_change_device=True,
    )

    # Restore zero.Init context if necessary
    zero.partition_parameters.restore_init_context()

    runner = ReaLDeepSpeedEngine(
        module=model,
        ds_engine=engine,
    )
    return_items = [
        runner,
        engine.optimizer,
        engine.training_dataloader,
        engine.lr_scheduler,
    ]
    return tuple(return_items)


@dataclasses.dataclass
class DeepspeedBackend(model_api.ModelBackend):
    optimizer_name: str = "adam"
    optimizer_config: dict = dataclasses.field(
        default_factory=lambda: dict(
            lr=1e-5, weight_decay=0.1, betas=(0.9, 0.95), eps=1e-5
        )
    )
    lr_scheduler_type: str = "cosine"
    warmup_steps_proportion: float = 0.0
    min_lr_ratio: float = 0.0  # will be used for linear and cosine schedule
    offload_param: bool = False
    offload_optimizer_state: bool = False
    enable_fp16: bool = True
    enable_bf16: bool = False
    zero_stage: int = 2
    # addtional deepspeed args
    additional_ds_config: Dict = dataclasses.field(default_factory=dict)

    def _initialize(self, model: model_api.Model, spec: model_api.FinetuneSpec):
        deepspeed.init_distributed(auto_mpi_discovery=False)
        module = model.module
        weight_decay = self.optimizer_config.get("weight_decay", 0.0)
        if self.optimizer_name == "adam":
            if not self.offload_param and not self.offload_optimizer_state:
                optim_cls = deepspeed.ops.adam.FusedAdam
            else:
                optim_cls = deepspeed.ops.adam.DeepSpeedCPUAdam
            optimizer = optim_cls(
                get_optimizer_grouped_parameters(module, weight_decay),
                **self.optimizer_config,
            )
        else:
            raise NotImplementedError(
                f"Unsupported optimizer: {self.optimizer_name}."
            )

        ds_config = get_train_ds_config(
            offload_param=self.offload_param,
            offload_optimizer_state=self.offload_optimizer_state,
            stage=self.zero_stage,
            enable_bf16=self.enable_bf16,
            enable_fp16=self.enable_fp16,
            **self.additional_ds_config,
        )

        # NOTE: Just a fake batch size to make DeepSpeed happy.
        ds_config["train_batch_size"] = constants.data_parallel_world_size()

        def warmup_then_cosine_anneal(
            step, warmup_steps_proportion, total_steps, min_lr_ratio
        ):
            warmup_steps = max(5, int(total_steps * warmup_steps_proportion))
            cosine_steps = total_steps - warmup_steps
            if step < warmup_steps:
                return 1.0 / warmup_steps * step
            return min_lr_ratio + 0.5 * (1.0 - min_lr_ratio) * (
                1 + math.cos((step - warmup_steps) / cosine_steps * math.pi)
            )

        def warmup_then_linear_anneal(
            step, warmup_steps_proportion, total_steps, min_lr_ratio
        ):
            warmup_steps = max(5, int(total_steps * warmup_steps_proportion))
            linear_steps = total_steps - warmup_steps
            if step < warmup_steps:
                return 1.0 / warmup_steps * step
            return 1.0 - (1.0 - min_lr_ratio) / linear_steps * (
                step - warmup_steps
            )

        def warmup_then_constant_anneal(
            step, warmup_steps_proportion, total_steps, min_lr_ratio
        ):
            warmup_steps = max(5, int(total_steps * warmup_steps_proportion))
            if step < warmup_steps:
                return 1.0 / warmup_steps * step
            return 1.0

        if self.lr_scheduler_type == "cosine":
            lr_scheduler_fn = warmup_then_cosine_anneal
        elif self.lr_scheduler_type == "linear":
            lr_scheduler_fn = warmup_then_linear_anneal
        elif self.lr_scheduler_type == "constant":
            lr_scheduler_fn = warmup_then_constant_anneal
        else:
            raise NotImplementedError(
                f"Unknown lr_scheduler_type {self.lr_scheduler_type}."
            )

        lr_lambda = functools.partial(
            lr_scheduler_fn,
            warmup_steps_proportion=self.warmup_steps_proportion,
            total_steps=spec.total_train_steps,
            min_lr_ratio=self.min_lr_ratio,
        )
        lr_scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda)

        module, *_ = deepspeed_initialize(
            model=module,
            optimizer=optimizer,
            config=ds_config,
            lr_scheduler=lr_scheduler,
        )

        model.module = module
        return model


model_api.register_backend("deepspeed", DeepspeedBackend)



from typing import Literal, Optional
import logging.config
import os

import colorlog

LOG_FORMAT = "%(asctime)s.%(msecs)03d %(name)s %(levelname)s: %(message)s"
DATE_FORMAT = "%Y%m%d-%H:%M:%S"
LOGLEVEL = logging.INFO

# NOTE: To use colorlog we should not call colorama.init() anywhere.
# The available color names are black, red, green, yellow, blue, purple, cyan and white
log_config = {
    "version": 1,
    "formatters": {
        "plain": {
            "()": colorlog.ColoredFormatter,
            "format": "%(log_color)s" + LOG_FORMAT,
            "datefmt": DATE_FORMAT,
            "log_colors": {
                "DEBUG": "white",
                "INFO": "white",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "bold_white,bg_red",
            },
        },
        "colored": {
            "()": colorlog.ColoredFormatter,
            "format": "%(log_color)s" + LOG_FORMAT,
            "datefmt": DATE_FORMAT,
            "log_colors": {
                "DEBUG": "blue",
                "INFO": "light_purple",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "bold_white,bg_red",
            },
        },
        "colored_system": {
            "()": colorlog.ColoredFormatter,
            "format": "%(log_color)s" + LOG_FORMAT,
            "datefmt": DATE_FORMAT,
            "log_colors": {
                "DEBUG": "blue",
                "INFO": "light_green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "bold_white,bg_red",
            },
        },
        "colored_benchmark": {
            "()": colorlog.ColoredFormatter,
            "format": "%(log_color)s" + LOG_FORMAT,
            "datefmt": DATE_FORMAT,
            "log_colors": {
                "DEBUG": "light_black",
                "INFO": "light_cyan",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "bold_white,bg_red",
            },
        },
    },
    "handlers": {
        "plainHandler": {
            "class": "logging.StreamHandler",
            "level": LOGLEVEL,
            "formatter": "plain",
            "stream": "ext://sys.stdout",
        },
        "benchmarkHandler": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "colored_benchmark",
            "stream": "ext://sys.stdout",
        },
        "systemHandler": {
            "class": "logging.StreamHandler",
            "level": "INFO",
            "formatter": "colored_system",
            "stream": "ext://sys.stdout",
        },
        "coloredHandler": {
            "class": "logging.StreamHandler",
            "level": LOGLEVEL,
            "formatter": "colored",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "plain": {
            "handlers": ["plainHandler"],
            "level": LOGLEVEL,
        },
        "benchmark": {
            "handlers": ["benchmarkHandler"],
            "level": "DEBUG",
        },
        "colored": {
            "handlers": ["coloredHandler"],
            "level": LOGLEVEL,
        },
        "system": {
            "handlers": ["systemHandler"],
            "level": LOGLEVEL,
        },
    },
}


def getLogger(
    name: Optional[str] = None,
    type_: Optional[Literal["plain", "benchmark", "colored", "system"]] = None,
):
    logging.config.dictConfig(log_config)
    if name is None:
        name = "plain"
    if type_ is None:
        type_ = "plain"
    assert type_ in ["plain", "benchmark", "colored", "system"]
    if name not in log_config["loggers"]:
        log_config["loggers"][name] = {
            "handlers": [f"{type_}Handler"],
            "level": LOGLEVEL,
        }
        logging.config.dictConfig(log_config)
    return logging.getLogger(name)


if __name__ == "__main__":
    # The following serves as a color visualization test.
    # The available color names are black, red, green, yellow, blue, purple, cyan and white
    log_config = {
        "version": 1,
        "formatters": {
            "colored": {
                "()": colorlog.ColoredFormatter,
                "format": "%(log_color)s" + LOG_FORMAT,
                "datefmt": DATE_FORMAT,
                "log_colors": {
                    "DEBUG": "purple",
                    "INFO": "light_purple",
                    "WARNING": "yellow",
                    "ERROR": "red",
                    "CRITICAL": "bold_white,bg_red",
                },
            },
        },
        "handlers": {
            "coloredHandler": {
                "class": "logging.StreamHandler",
                "level": "DEBUG",
                "formatter": "colored",
                "stream": "ext://sys.stdout",
            },
        },
        "loggers": {
            "": {
                "handlers": ["coloredHandler"],
                "level": "DEBUG",
            },
        },
    }
    logging.config.dictConfig(log_config)
    logging.debug("This is a debug message")
    logging.info("This is an info message")
    logging.warning("This is a warning message")
    logging.error("This is an error message")
    logging.critical("This is a critical message")



from collections import defaultdict
from typing import *
import dataclasses
import itertools
import os
import socket

import torch.distributed

from realhf.api.core import system_api
from realhf.api.core.config import ModelName
from realhf.base import (
    constants,
    gpu_utils,
    name_resolve,
    names,
    network,
    topology,
)


@dataclasses.dataclass
class NCCLProcessGroupInfo:
    world_size: int
    global_rank: int
    local_gpu_id: int
    # 3D parallelism groups of each model.
    model_groups: Dict[str, torch.distributed.ProcessGroup]


def filter_match_mwids(
    model_name: ModelName,
    topo: topology.PipeModelDataParallelTopology,
    msid2mwid: Dict[system_api.ModelShardID, int],
    **conditions,
) -> List[int]:
    if len(conditions) == 0:
        mwids_this_model = [
            msid2mwid[
                system_api.ModelShardID.from_parallelism_rank(
                    model_name, topo, j
                )
            ]
            for j in range(topo.world_size())
        ]
    else:
        mwids_this_model = [
            msid2mwid[
                system_api.ModelShardID.from_parallelism_rank(
                    model_name, topo, j
                )
            ]
            for j in topo.filter_match(**conditions)
        ]
    mwids_this_model = sorted(mwids_this_model)
    assert len(set(mwids_this_model)) == len(mwids_this_model)
    return list(mwids_this_model)


def setup_global_comm(
    expr_name: str,
    trial_name: str,
    worker_index: int,
    model_topos: Optional[
        Dict[str, topology.PipeModelDataParallelTopology]
    ] = None,
    msid2mwid: Optional[Dict[system_api.ModelShardID, int]] = None,
    world_size: Optional[int] = None,  # for testing only
    global_rank: Optional[int] = None,  # for testing only
) -> NCCLProcessGroupInfo:
    assert (world_size is None) == (global_rank is None)
    if world_size is None:
        peers: List[int] = list(
            sorted(
                map(
                    int,
                    name_resolve.get_subtree(
                        names.trainer_ddp_peer(
                            expr_name,
                            trial_name,
                            gpu_utils.GLOBAL_PROCESS_GROUP_NAME,
                        )
                    ),
                )
            )
        )
        assert len(peers) == len(
            set(peers)
        ), f"Duplicated trainer worker index. {peers}"
        world_size = len(peers)
        global_rank = peers.index(worker_index)

    mw_ranks = {}
    if model_topos is not None:
        assert msid2mwid is not None
        for model_name, topo in model_topos.items():
            mw_ranks[model_name] = filter_match_mwids(
                model_name, topo, msid2mwid
            )

    if (
        "GPU_DEVICES_ISOLATED" not in os.environ
        and "RAY" not in os.environ["REAL_MODE"]
    ):
        raise RuntimeError(
            "GPU devices not isolated in slurm or local mode. This should not happen."
        )

    assert len(os.environ["CUDA_VISIBLE_DEVICES"].split(",")) == 1, os.environ[
        "CUDA_VISIBLE_DEVICES"
    ]
    local_gpu_id = int(os.environ["CUDA_VISIBLE_DEVICES"])

    ddp_master_name = names.trainer_ddp_master(
        expr_name, trial_name, gpu_utils.GLOBAL_PROCESS_GROUP_NAME
    )

    if worker_index == 0:
        host_ip = socket.gethostbyname(socket.gethostname())
        port = network.find_free_port()
        ddp_init_address = f"tcp://{host_ip}:{port}"
        name_resolve.add(ddp_master_name, ddp_init_address, keepalive_ttl=300)
    else:
        try:
            ddp_init_address = name_resolve.wait(ddp_master_name, timeout=300)
        except TimeoutError:
            raise TimeoutError(
                f"global_rank={global_rank} worker_index={worker_index} wait for ddp_init_method timeout."
            )

    torch_dist_kwargs = dict(
        world_size=world_size,
        rank=global_rank,
        init_method=ddp_init_address,
        backend="nccl",
    )
    torch.cuda.set_device(
        0
    )  # initialize CUDA here with only a single visible device
    # This environment variable is used by DeepSpeed.
    os.environ["LOCAL_RANK"] = "0"

    torch.distributed.init_process_group(
        **torch_dist_kwargs, group_name=gpu_utils.GLOBAL_PROCESS_GROUP_NAME
    )

    model_groups = {}
    for model_name, ranks in mw_ranks.items():
        model_groups[model_name] = topology.new_or_get_group(
            ranks, backend="nccl"
        )
        constants.set_parallelism_group(
            model_name, model_groups[model_name], ranks
        )

    self_group = None
    for i in range(world_size):
        group = topology.new_or_get_group([i], backend="nccl")
        if i == global_rank:
            self_group = group
            constants.set_self_group(self_group)

    # logger.info(f"Setup process group finishes for worker_index={worker_index}")

    return NCCLProcessGroupInfo(
        world_size=world_size,
        global_rank=global_rank,
        local_gpu_id=local_gpu_id,
        model_groups=model_groups,
    )



from typing import Any, Dict, List, Literal, Optional
import argparse
import functools
import json
import os
import pickle
import re

import numpy as np

try:
    import realhf._C.mdm_search as mdm_search
except ModuleNotFoundError:
    mdm_search = None

from realhf.api.core.dfg import MFCDef, ModelInterfaceType
from realhf.api.core.model_api import ModelFamily
from realhf.api.quickstart.device_mesh import DeviceMesh, RPCAllocation
from realhf.api.quickstart.model import ModelTrainEvalConfig, ParallelismConfig
from realhf.api.quickstart.search import RPCExecution
import realhf.base.constants as constants


def search_rpc_allocations(
    device_mesh: DeviceMesh,
    rpcs: List[MFCDef],
    num_gen_tokens: int = 256,
    n_ppo_minibatches: int = 1,
    seq_len: int = 256,
    gradient_checkpointing: bool = True,
    use_cache: bool = False,
) -> List[RPCAllocation]:
    from realhf.search_engine.enumerate import build_graph
    from realhf.search_engine.estimate import get_param_realloc_stats

    from_file = os.environ.get("IS_REMOTE", "0") == "1"
    dump_dir = os.path.join(
        constants.LOG_ROOT,
        constants.experiment_name(),
        constants.trial_name(),
        "device_mapping.pkl",
    )
    log_dir = os.path.join(
        constants.LOG_ROOT,
        constants.experiment_name(),
        constants.trial_name(),
        "device_mapping",
    )
    rs_dir = os.path.join(
        constants.LOG_ROOT,
        constants.experiment_name(),
        constants.trial_name(),
        "raw_search_result",
    )
    rpc_exe_dir = os.path.join(
        constants.LOG_ROOT,
        constants.experiment_name(),
        constants.trial_name(),
        "rpc_exe_info",
    )

    if from_file or (use_cache and os.path.exists(dump_dir)):
        with open(dump_dir, "r") as f:
            s = json.load(f)
            rpc_allocs = [RPCAllocation.from_dict(d) for d in s]
        return rpc_allocs
    else:
        os.makedirs(os.path.dirname(dump_dir), exist_ok=True)

    n_nodes = device_mesh.n_nodes
    table = {}
    for rpc in rpcs:
        print(
            f"Getting param realloc stats for {rpc.model_type} at {rpc.model_path}"
        )
        t = get_param_realloc_stats(
            rpc.model_type, rpc.model_path, n_nodes, True
        )
        table.update(t)

    rpc_exe_list = make_rpc_exe_list(
        rpcs,
        device_mesh,
        num_gen_tokens=num_gen_tokens,
        n_ppo_minibatches=n_ppo_minibatches,
        seq_len=seq_len,
        gradient_checkpointing=gradient_checkpointing,
        log_dir=rpc_exe_dir,
        if_print=False,
    )
    graph = build_graph(rpcs, 5, 1, if_print=False)
    model_size_dict = make_model_size_dict(rpcs, if_print=False)

    n_nodes = device_mesh.n_nodes
    search_time = 120

    rs: List[Dict[str, List]] = mdm_search.multi_mcmc_search(
        rpcs,
        rpc_exe_list,
        graph,
        table,
        model_size_dict,
        0.001,  # beta min
        0.002,  # beta max
        0.001,  # beta step
        search_time,  # time limit for each search
        1,  # repeat
    )
    if not from_file:
        with open(rs_dir, "w") as f:
            import pprint

            pprint.pprint(rs, stream=f)

    import pprint

    r: Dict[str, Dict[str, Any]] = rs[-1]
    pprint.pprint(r)

    rpc_name_to_rpcs = {rpc.name: rpc for rpc in rpcs}
    rpc_allocs = []
    for rpc_name, alloc_info in r.items():
        if rpc_name in ["end_time", "mem_cost"]:
            continue
        # rpc = rpc_dict[rpc_name]
        rpc = rpc_name_to_rpcs[rpc_name]
        parallel = ParallelismConfig(
            pipeline_parallel_size=alloc_info["num_pp"],
            data_parallel_size=alloc_info["num_dp"],
            model_parallel_size=alloc_info["num_mp"],
            use_sequence_parallel=(
                alloc_info["num_mp"] > 1
                and rpc.interface_type == ModelInterfaceType.TRAIN_STEP
            ),
        )
        sub_device_mesh = DeviceMesh(
            n_nodes=device_mesh.n_nodes,
            n_gpus_per_node=device_mesh.n_gpus_per_node,
            mapping=alloc_info["device_mesh_mapping"],
            name=alloc_info["device_mesh_name"],
            global_mesh_name=device_mesh.global_mesh_name,
        )
        rpc_alloc = RPCAllocation(
            rpc=rpc,
            device_mesh=sub_device_mesh,
            parallel=parallel,
        )
        rpc_allocs.append(rpc_alloc)

    if not from_file:
        with open(dump_dir, "w") as f:
            json.dump(
                [rpc_alloc.to_dict() for rpc_alloc in rpc_allocs], f, indent=4
            )
        with open(log_dir, "w") as f:
            import pprint

            pprint.pprint(rpc_allocs, stream=f)

    return rpc_allocs


def make_rpc_exe_list(
    rpcs: List[MFCDef],
    device_mesh: DeviceMesh,
    num_gen_tokens: int,
    n_ppo_minibatches: int,
    seq_len: int,
    gradient_checkpointing: bool,
    if_print: bool = False,
    log_dir: Optional[str] = None,
) -> List[RPCExecution]:
    from realhf.search_engine.enumerate import enumerate_rpc_executions

    rpc_exe_list = []
    log_flag = False
    for rpc in rpcs:
        # real_model_config = load_model_config(rpc)
        feasible = enumerate_rpc_executions(
            rpc,
            device_mesh,
            seq_len=seq_len,
            num_gen_tokens=num_gen_tokens,
            n_ppo_minibatches=n_ppo_minibatches,
            gradient_checkpointing=gradient_checkpointing,
        )
        rpc_exe_list.extend(feasible)

        if log_dir is not None:
            mode = "w" if not log_flag else "a"
            with open(log_dir, mode) as f:
                f.write(f"{rpc.name} feasible: {len(feasible)}\n")
                feasible.sort(key=lambda x: x.time_cost)
                # feasible = feasible[:30]
                for i, rpc_exe in enumerate(feasible):
                    f.write(
                        f"{i}: time_cost: {rpc_exe.time_cost} ms, {rpc_exe.time_cost} "
                        f"sub_device_mesh: {rpc_exe.device_mesh}, "
                        f"parallel_strategy: {rpc_exe.parallel_strategy}, "
                        f"mem_cost: {rpc_exe.mem/(1024*1024*1024):02f} GB, "
                        f"static_mem_cost: {rpc_exe.static_mem/(1024*1024*1024):02f} GB\n"
                    )
                f.write("\n")
                log_flag = True

        if if_print:
            print(f"{rpc.name} feasible: {len(feasible)}")
            feasible.sort(key=lambda x: x.time_cost)
            # feasible = feasible[:10]
            for i, rpc_exe in enumerate(feasible):
                print(
                    f"{i}: time_cost: {rpc_exe.time_cost} ms, "
                    f"sub_device_mesh: {rpc_exe.device_mesh}, "
                    f"parallel_strategy: {rpc_exe.parallel_strategy}, "
                    f"mem_cost: {rpc_exe.mem/(1024*1024*1024):02f} GB, "
                    f"static_mem_cost: {rpc_exe.static_mem/(1024*1024*1024):02f} GB"
                )

    return rpc_exe_list


def make_model_size_dict(
    rpcs: List[MFCDef], if_print: bool = False
) -> Dict[str, int]:
    model_size_dict = {}

    for rpc in rpcs:
        if rpc.model_name.role in model_size_dict:
            continue
        # model_configs = load_model_config(rpc)
        # model_size_dict[rpc.model_name.role] = estimate_model_size(real_model_config)
        model_size_dict[rpc.model_name.role] = rpc.model_type.size

        if if_print:
            print(
                f"model_name: {rpc.model_name.role}, "
                f"model_size: {rpc.model_type.size}"
            )
    return model_size_dict



# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.

# Parts of the code here are adapted from PyTorch
# repo: https://github.com/pytorch/pytorch

# adopted from megatron

import contextlib

from torch import _C
from torch.cuda import _lazy_call
from torch.cuda import device as device_ctx_manager
from torch.utils.checkpoint import detach_variable
import torch

from realhf.impl.model.parallelism.model_parallel.utils import (
    gather_split_1d_tensor,
    safely_set_viewless_tensor_data,
    split_tensor_into_1d_equal_chunks,
)
import realhf.base.constants as constants

# Default name for the model parallel rng tracker.
_MODEL_PARALLEL_RNG_TRACKER_NAME = "model-parallel-rng"
_EXPERT_PARALLEL_RNG_TRACKER_NAME = "expert-parallel-rng"
_DATA_PARALLEL_RNG_TRACKER_NAME = "data-parallel-rng"


def _set_cuda_rng_state(new_state, device=-1):
    """Sets the random number generator state of the current GPU.

    Argumentss:
        new_state (torch.ByteTensor): The desired state
    This function is adapted from PyTorch repo (torch.cuda.set_rng_state)
    with a single change: the input state is not cloned. Cloning caused
    major performance issues for +4 GPU cases.
    """
    if hasattr(_C, "_cuda_setRNGState") and callable(_C._cuda_setRNGState):
        # older PyTorch
        def cb():
            with device_ctx_manager(device):
                _C._cuda_setRNGState(new_state)

    else:
        # newer PyTorch
        if device == -1:
            device = torch.device("cuda")
        elif isinstance(device, str):
            device = torch.device(device)
        elif isinstance(device, int):
            device = torch.device("cuda", device)

        def cb():
            idx = device.index
            if idx is None:
                idx = torch.cuda.current_device()
            default_generator = torch.cuda.default_generators[idx]
            default_generator.set_state(new_state)

    _lazy_call(cb)


def get_expert_parallel_rng_tracker_name():
    global _EXPERT_PARALLEL_RNG_TRACKER_NAME
    return _EXPERT_PARALLEL_RNG_TRACKER_NAME


def get_data_parallel_rng_tracker_name():
    global _DATA_PARALLEL_RNG_TRACKER_NAME
    return _DATA_PARALLEL_RNG_TRACKER_NAME


class CudaRNGStatesTracker:
    """Tracker for the cuda RNG states.

    Using the `add` method, a cuda rng state is initialized based on
    the input `seed` and is assigned to `name`. Later, by forking the
    rng state, we can perform operations and return to our starting
    cuda state.
    """

    def __init__(self):
        # Map from a string name to the cuda rng state.
        self.states_ = {}
        # Seeds are just for book keeping and ensure no seed is set twice.
        self.seeds_ = set()

    def reset(self):
        """Set to the initial state (no tracker)."""
        self.states_ = {}
        self.seeds_ = set()

    def get_states(self):
        """Get rng states. Copy the dictionary so we have direct
        pointers to the states, not just a pointer to the dictionary."""
        states = {}
        for name in self.states_:
            states[name] = self.states_[name]
        return states

    def set_states(self, states):
        """Set the rng states. For efficiency purposes, we do not check
        the size of seed for compatibility."""
        self.states_ = states

    def add(self, name, seed):
        """Track the rng state."""
        # Check seed is not already used.
        if seed in self.seeds_:
            raise Exception("seed {} already exists".format(seed))
        self.seeds_.add(seed)
        # Check that state is not already defined.
        if name in self.states_:
            raise Exception("cuda rng state {} already exists".format(name))
        # Get the current rng state.
        orig_rng_state = torch.cuda.get_rng_state()
        # Set the new state and store it.
        torch.cuda.manual_seed(seed)
        self.states_[name] = torch.cuda.get_rng_state()
        # Reset rng state to what it was.
        _set_cuda_rng_state(orig_rng_state)

    @contextlib.contextmanager
    def fork(self, name=_MODEL_PARALLEL_RNG_TRACKER_NAME):
        """Fork the cuda rng state, perform operations, and exit with
        the original state."""
        # Check if we have added the state
        if name not in self.states_:
            raise Exception("cuda rng state {} is not added".format(name))
        # Store current rng state.
        orig_cuda_rng_state = torch.cuda.get_rng_state()
        # Set rng state to the desired one
        _set_cuda_rng_state(self.states_[name])
        # Do the stuff we wanted to do.
        try:
            yield
        finally:
            # Update the current rng state for later use.
            self.states_[name] = torch.cuda.get_rng_state()
            # And set the state to the original state we started with.
            _set_cuda_rng_state(orig_cuda_rng_state)


# RNG tracker object.
_CUDA_RNG_STATE_TRACKER = CudaRNGStatesTracker()


def get_cuda_rng_tracker():
    """Get cuda rng tracker."""
    return _CUDA_RNG_STATE_TRACKER


def model_parallel_cuda_manual_seed(seed):
    """Initialize model parallel cuda seed.

    This function should be called after the model parallel is
    initialized. Also, no torch.cuda.manual_seed should be called
    after this function. Basically, this is replacement for that
    function.
    Two set of RNG states are tracked:
    default state: This is for data parallelism and is the same among a set of model parallel GPUs but different across different model paralle groups. This is used for example for dropout in the non-tensor-model-parallel regions.
    tensor-model-parallel state: This state is different among a set of model parallel GPUs, but the same across data parallel groups. This is used for example for dropout in model parallel regions.
    """
    # 2718 is just for fun and any POSITIVE value will work.
    model_parallel_rank = constants.model_parallel_rank()
    offset = seed + 2718
    tensor_model_parallel_seed = offset + model_parallel_rank
    # Data parallel gets the original seed.
    data_parallel_seed = seed

    _CUDA_RNG_STATE_TRACKER.reset()
    # Set the default state.
    torch.cuda.manual_seed(data_parallel_seed)
    _CUDA_RNG_STATE_TRACKER.add(
        _DATA_PARALLEL_RNG_TRACKER_NAME, data_parallel_seed
    )

    # and model parallel state.
    _CUDA_RNG_STATE_TRACKER.add(
        _MODEL_PARALLEL_RNG_TRACKER_NAME, tensor_model_parallel_seed
    )

    expert_parallel_seed = seed + 1024 + model_parallel_rank
    _CUDA_RNG_STATE_TRACKER.add(
        _EXPERT_PARALLEL_RNG_TRACKER_NAME, expert_parallel_seed
    )


class CheckpointFunction(torch.autograd.Function):
    """Checkpoint Function

    This function is adapted from torch.utils.checkpoint with two main changes:
    1) torch.cuda.set_rng_state is replaced with `_set_cuda_rng_state`
    2) the states in the model parallel tracker are also properly tracked/set/reset.
    """

    @staticmethod
    def forward(ctx, run_function, distribute_saved_activations, *args):
        ctx.run_function = run_function
        ctx.distribute_saved_activations = distribute_saved_activations

        # Copy the rng states.
        ctx.fwd_cpu_rng_state = torch.get_rng_state()
        ctx.fwd_cuda_rng_state = torch.cuda.get_rng_state()
        ctx.fwd_cuda_rng_state_tracker = get_cuda_rng_tracker().get_states()

        with torch.no_grad():
            outputs = run_function(*args)

        # Divide hidden states across model parallel group and only keep
        # the chunk corresponding to the current rank.
        if distribute_saved_activations:
            ctx.input_0_shape = args[0].data.shape
            safely_set_viewless_tensor_data(
                args[0],
                split_tensor_into_1d_equal_chunks(
                    args[0].data, new_buffer=True
                ),
            )

        # Store everything.
        ctx.save_for_backward(*args)

        return outputs

    @staticmethod
    def backward(ctx, *args):
        if not torch.autograd._is_checkpoint_valid():
            raise RuntimeError(
                "Checkpointing is not compatible with .grad(), "
                "please use .backward() if possible"
            )
        inputs = ctx.saved_tensors
        if ctx.distribute_saved_activations:
            safely_set_viewless_tensor_data(
                inputs[0],
                gather_split_1d_tensor(inputs[0].data).view(ctx.input_0_shape),
            )

        # Store the current states.
        bwd_cpu_rng_state = torch.get_rng_state()
        bwd_cuda_rng_state = torch.cuda.get_rng_state()
        bwd_cuda_rng_state_tracker = get_cuda_rng_tracker().get_states()

        # Set the states to what it used to be before the forward pass.
        torch.set_rng_state(ctx.fwd_cpu_rng_state)
        _set_cuda_rng_state(ctx.fwd_cuda_rng_state)
        get_cuda_rng_tracker().set_states(ctx.fwd_cuda_rng_state_tracker)

        # Compute the forward pass.
        detached_inputs = detach_variable(inputs)
        with torch.enable_grad():
            outputs = ctx.run_function(*detached_inputs)

        # Set the states back to what it was at the start of this function.
        torch.set_rng_state(bwd_cpu_rng_state)
        _set_cuda_rng_state(bwd_cuda_rng_state)
        get_cuda_rng_tracker().set_states(bwd_cuda_rng_state_tracker)

        if isinstance(outputs, torch.Tensor):
            outputs = (outputs,)

        # filter out non tensor outputs for backward pass
        outputs, args = zip(
            *filter(lambda x: torch.is_tensor(x[0]), zip(outputs, args))
        )
        torch.autograd.backward(outputs, args)
        grads = tuple(
            inp.grad if isinstance(inp, torch.Tensor) else inp
            for inp in detached_inputs
        )
        return (None, None) + grads


def checkpoint(function, distribute_saved_activations, *args):
    """Checkpoint a model or part of the model.
    This has been directly copied from torch.utils.checkpoint."""
    return CheckpointFunction.apply(
        function, distribute_saved_activations, *args
    )
